import logging
import sys
from collections import namedtuple, defaultdict
from pathlib import Path
from typing import Union, List, Dict, Type, Set, Tuple, Optional

import inspect

import numpy as np
from astropy.io import fits
from astropy.io.fits.hdu.base import _BaseHDU
from astropy.table import Table, Column

from weaveio.file import File, PrimaryHDU, TableHDU
from weaveio.graph import Graph
from weaveio.hierarchy import Multiple, unwind, collect, Hierarchy, find_branch
from weaveio.opr3.hierarchy import APS, OB, OBSpec, Exposure, WeaveTarget, _predicate, Run, ArmConfig, FibreTarget, Fibre
from weaveio.opr3.l1 import L1Spectrum, L1SingleSpectrum, L1StackSpectrum, L1SupertargetSpectrum
from weaveio.opr3.l2 import L2, L2Single, L2Stack, L2Superstack, L2Supertarget, IngestedSpectrum, Fit, ModelSpectrum, Redrock, \
    RVSpecfit, Ferre, PPXF, Gandalf, GandalfModelSpectrum, CombinedIngestedSpectrum, CombinedModelSpectrum, Template, RedshiftArray, GandalfEmissionModelSpectrum, GandalfCleanModelSpectrum, \
    GandalfCleanIngestedSpectrum, L2Product, gandalf_line_names, gandalf_index_names, UncombinedIngestedSpectrum, UncombinedModelSpectrum, BaseCombinedModelSpectrum
from weaveio.opr3.l1files import L1File, L1SuperstackFile, L1StackFile, L1SingleFile, L1SupertargetFile
from weaveio.writequery import CypherData, CypherVariable
from weaveio.writequery.actions import string_append
from weaveio.writequery.base import CypherAppendStr


MAX_REDSHIFT_GRID_LENGTH = 5000

class MissingDataError(Exception):
    pass


def column_name_acronym_replacements(columns:List[str], *acronyms: str) -> Dict[str, str]:
    replacements = {}
    for column in columns:
        for a in acronyms:
            for aa in [f'_{a}', f'_{a}_', f'{a}_']:
                if aa in column:
                    replacements[column.replace(aa.lower(), '')] = column
    return replacements


def filter_products_from_table(table: Table, maxlength: int) -> Table:
    columns = []
    for i in table.colnames:
        value = table[i]
        if len(value.shape) == 2:
            if value.shape[1] > maxlength:
                continue
        columns.append(i)
    t = table[columns]
    for col in table.colnames:
        t.rename_column(col, col.lower())
    return t


def extract_lines_names(expected_line_names, colnames):
    """
    returns dictionary of required factor names and values that can be take from the hdu
    """
    colnames = list(map(lambda x: x.lower(), colnames))
    expected_line_names = list(map(lambda x: x.lower(), expected_line_names))
    actual_line_names = sorted([i[len('FLUX') + 1:] for i in colnames if i.startswith('flux')], key=lambda x: float(x.split('_')[1]))
    d = {}
    for name in actual_line_names:
        if name in expected_line_names:
            d[f"{name}_aon"] = f'aon_{name}'
            d[f"{name}_ebmv"] = f'ebv_{name}'
            d[f"{name}_flux"] = f'flux_{name}'
            d[f"{name}_flux_error"] = f'flux_err_{name}'
            d[f"{name}_amp_error"] = f'ampl_err_{name}'
            d[f"{name}_redshift"] = f'z_{name}'
            d[f"{name}_redshift_error"] = f'z_err_{name}'
            d[f"{name}_sigma"] = f'sigma_{name}'
            d[f"{name}_sigma_error"] = f'sigma_err_{name}'
        else:
            logging.warning(f"line `{name}` was found in file but is not part of the schema (will be ignored)")
    for name in expected_line_names:
        if name not in actual_line_names:
            logging.warning(f"expected line `{name}` not found in l2file[gandalf] hdu")
    return d


def extract_indices_names(expected_index_names, colnames):
    d = {}
    for name in expected_index_names:
        name = name.lower()
        if name in colnames:
            d[f'{name}'] = name
            d[f'{name}_error'] = f'err_{name}'
        else:
            logging.warning(f"index `{name}` was not found in l2file[gandalf] hdu")
    return d

FitSpecs = namedtuple('FitSpecs', ['individuals', 'individual_models', 'combined', 'combined_model', 'colour_codes', 'nrow'])
GandalfSpecs = namedtuple('GandalfSpecs', ['model', 'ingested', 'emission', 'clean_model', 'clean_ingested', 'nrow'])


class L2File(File):
    singular_name = 'l2file'
    is_template = True
    match_pattern = '.*APS.fits'
    antimatch_pattern = '.*cube.*'
    L2 = L2Product
    L1 = L1Spectrum
    parents = [Multiple(L1File, 2, 3), Multiple(L2, maxnumber=1000, one2one=True)]
    children = [APS]
    parts = ['RR', 'RVS', 'FR', 'GAND', 'PPXF']
    recommended_batchsize = 50
    hdus = {
        'primary': PrimaryHDU,
        'class_table': TableHDU,
        'stellar_table': TableHDU,
        'galaxy_table': TableHDU,
        'class_spectra': TableHDU,
        'stellar_spectra': TableHDU,
        'galaxy_spectra': TableHDU,
            }

    @classmethod
    def length(cls, path, part=None):
        hdus = fits.open(path)
        if part is None:
            return hdus[1].header['NAXIS2']
        d = {'RR': 1, 'RVS': 2, 'FR':2, 'GAND': 3, 'PPXF': 3}
        return hdus[d[part]].header['NAXIS2']

    @classmethod
    def make_l2(cls, spectra, nspec, **hiers):
        return cls.L2(**{cls.L1.plural_name: [spectra[i] for i in range(nspec)]}, **hiers)

    @classmethod
    def decide_filetype(cls, l1filetypes: List[Type[File]]) -> Type[File]:
        l1precedence = [L1SingleFile, L1StackFile, L1SuperstackFile, L1SupertargetFile]
        l2precedence = [L2SingleFile, L2StackFile, L2SuperstackFile, L2SupertargetFile]
        highest = max(l1precedence.index(l1filetype) for l1filetype in l1filetypes)
        return l2precedence[highest]

    @classmethod
    def match_file(cls, directory: Union[Path, str], fname: Union[Path, str], graph: Graph):
        """
        L2 files can be formed from any combination of L1 files and so the shared hierarchy level can be
        either exposure, OB, OBSpec, or WeaveTarget.
        L2 files are distinguished by the shared hierarchy level of their formative L1 files.
        Therefore, we assign an L2 file to the highest hierarchy level.
        e.g.
        L1Single+L1Single -> L2Single
        L1Stack+L1Single -> L2Stack
        L1SuperStack+L1Stack -> L2SuperStack
        """
        fname = Path(fname)
        directory = Path(directory)
        path = directory / fname
        if not super().match_file(directory, fname, graph):
            return False
        header = cls.read_header_and_aps(path)[0]
        ftypes, _ = zip(*cls.parse_fname(header, fname, instantiate=False))
        return cls.decide_filetype(ftypes) is cls

    @classmethod
    def parser_ftypes_runids(cls, header):
        header_info = [header.get(f'L1_REF_{i}', '.').split('.')[0].split('_') for i in range(4)]
        ftypes_header, runids_header = zip(*[i for i in header_info if len(i) > 1])
        return ftypes_header, list(map(int, runids_header))

    @classmethod
    def parse_fname(cls, header, fname, instantiate=True) -> List[L1File]:
        """
        Return the L1File type and the expected filename that formed this l2 file
        """
        ftype_dict = {
            'single': L1SingleFile,
            'stacked': L1StackFile, 'stack': L1StackFile,
            'superstack': L1SuperstackFile, 'superstacked': L1SuperstackFile
        }
        split = fname.name.lower().replace('aps.fits', '').replace('aps.fit', '').strip('_.').split('__')
        runids = []
        ftypes = []
        for i in split:
            ftype, runid = i.split('_')
            runids.append(int(runid))
            ftypes.append(str(ftype))
        if len(ftypes) == 1:
            ftypes = [ftypes[0]] * len(runids)  # they all have the same type if there is only one mentioned
        assert len(ftypes) == len(runids), "error parsing runids/types from fname"
        ftypes_header, runids_header = cls.parser_ftypes_runids(header)
        if not all(map(lambda x: x[0] == x[1], zip(runids, runids_header))):
            raise ValueError(f"There is a mismatch between runids in the filename and in in the header")
        if not all(map(lambda x: x[0] == x[1], zip(ftypes, ftypes_header))):
            raise ValueError(f"There is a mismatch between stack/single filetype in the filename and in in the header")
        files = []
        for ftype, runid in zip(ftypes, runids):
            ftype_cls = ftype_dict[ftype]
            fname = ftype_cls.fname_from_runid(runid)
            if instantiate:
                files.append(ftype_cls.find(fname=fname))
            else:
                files.append((ftype_cls, fname))
        return files

    @classmethod
    def find_shared_hierarchy(cls, path: Path) -> Dict:
        raise NotImplementedError

    @classmethod
    def read_header_and_aps(cls, path):
        return fits.open(path)[0].header, fits.open(path)[1].header['APS_V'].strip()

    @classmethod
    def read_hdus(cls, directory: Union[Path, str], fname: Union[Path, str], l1files: List[L1File],
                  **hierarchies: Union[Hierarchy, List[Hierarchy]]) -> Tuple[Dict[int, 'HDU'], 'File', List[_BaseHDU]]:
        fdict = {p.plural_name: [] for p in cls.parents if isinstance(p, Multiple) and issubclass(p.node, L1File)} # parse the 1lfile types separately
        for f in l1files:
            fdict[f.plural_name].append(f)
        hierarchies.update(fdict)
        hierarchies[cls.L2.plural_name] = hierarchies['l2']
        del hierarchies['l2']
        return super().read_hdus(directory, fname, **hierarchies)

    @classmethod
    def get_l1_filenames(cls, header):
        return [v for k, v in header.items() if 'APS_REF' in k]

    @classmethod
    def get_all_fibreids(cls, path):
        aps_ids = set()
        for hdu in fits.open(path)[1:]:
            try:
                aps_ids |= set(hdu.data['APS_ID'].tolist())
            except KeyError:
                pass
        return sorted(aps_ids)

    @classmethod
    def attach_products_to_spectra(cls, specs: FitSpecs, formatter, hdu, names: Dict[str, str], types, combined_suffix='_C'):
        UncombinedIngestedSpectrum, CombinedIngestedSpectrum, ModelSpectrum, CombinedModelSpectrum = types
        # this is less complicated than it looks
        # `specs` is a namedtuple containing ingested and model spectra collections ([fibre1_spectrum, fibre2_spectrum, ...])
        # in the case of `individual_models` and `individuals`, they are a collection of collections:
        #    [[fibre1_red_spectrum, fibre2_red_spectrum], [fibre1_blue_spectrum, fibre2_blue_spectrum], ...]
        # so we unwind those ones twice
        if specs.individual_models is not None:
            with unwind(specs.individual_models, specs.colour_codes, specs.nrow) as (spectra, colour_codes, nrow):
                with unwind(spectra, colour_codes) as (spectrum, colour_code):
                    column_name = string_append(f'MODEL_{formatter}_'.upper(), colour_code)
                    spectrum = ModelSpectrum.from_cypher_variable(spectrum)
                    spectrum.attach_product('flux', hdu, column_name=column_name, index=nrow)
                collect(spectrum)  # collect to avoid duplication of effort later
            collect(nrow)  # collect to avoid duplication of effort later
        if specs.individuals is not None:
            with unwind(specs.individuals, specs.colour_codes, specs.nrow) as (spectra, colour_codes, nrow):
                with unwind(spectra, colour_codes) as (spectrum, colour_code):
                    spectrum = UncombinedIngestedSpectrum.from_cypher_variable(spectrum)
                    for product in spectrum.products:
                        column_name = string_append(f'{names.get(product, product)}_{formatter}_'.upper(), colour_code)
                        spectrum.attach_product(product, hdu, column_name=column_name, index=nrow)
                collect(spectrum)  # collect to avoid duplication of effort later
            collect(nrow)  # collect to avoid duplication of effort later
        if specs.combined_model is not None:
            with unwind(specs.combined_model, specs.nrow) as (combined_model, nrow):
                spectrum = CombinedModelSpectrum.from_cypher_variable(combined_model)
                column_name = f'model_{formatter}{combined_suffix}'.upper()
                spectrum.attach_product('flux', hdu, column_name=column_name, index=nrow)
            collect(spectrum)  # collect to avoid duplication of effort later
        if specs.combined is not None:
            with unwind(specs.combined, specs.nrow) as (combined, nrow):
                spectrum = CombinedIngestedSpectrum.from_cypher_variable(combined)
                for product in spectrum.products:
                    column_name = f'{names.get(product, product)}_{formatter}{combined_suffix}'.upper()
                    spectrum.attach_product(product, hdu, column_name=column_name, index=nrow)
            collect(spectrum)  # collect to avoid duplication of effort later

    @classmethod
    def attach_products_to_gandalf_extra_spectra(cls, specs: GandalfSpecs, hdu):
        with unwind(specs.emission, specs.clean_model, specs.clean_ingested, specs.nrow) as (emission, clean_model, clean_ingested, nrow):
            emission = GandalfEmissionModelSpectrum.from_cypher_variable(emission)
            clean_model = GandalfCleanModelSpectrum.from_cypher_variable(clean_model)
            clean_ingested = GandalfCleanIngestedSpectrum.from_cypher_variable(clean_ingested)
            emission.attach_product('flux', hdu, column_name=f'EMISSION_GAND', index=nrow)
            clean_model.attach_product('flux', hdu, column_name=f'MODEL_CLEAN_GAND', index=nrow)
            clean_ingested.attach_product('flux', hdu, column_name=f'FLUX_CLEAN_GAND', index=nrow)
        collect(nrow)  # collect to avoid duplication of effort later

    @classmethod
    def make_redrock_fit(cls, l1spectra, specs, zs, row, nl1specs, replacements):
        templates = {}
        unrolled = [specs.individual_models[i] for i in range(nl1specs)]
        for template_name in Redrock.template_names:
            chi2s = row[f'CZZ_CHI2_{template_name}'.lower()]
            template = Template(l1_spectra=l1spectra, uncombined_model_spectra=unrolled, combined_model_spectrum=specs.combined_model,
                                redshift_array=zs[template_name], name=template_name, chi2_array=chi2s)
            templates[template_name] = template
        return Redrock(l1_spectra=l1spectra, uncombined_model_spectra=unrolled, combined_model_spectrum=specs.combined_model, tables=row, tables_replace=replacements, **templates)

    @classmethod
    def make_gandalf_structure(cls, l1spectra, specs, row, this_fname, nrow, replacements):
        model, ingested = specs.combined_model, specs.combined
        emission = GandalfEmissionModelSpectrum(gandalf_model_spectrum=model)
        clean_model = GandalfCleanModelSpectrum(gandalf_model_spectrum=model)
        clean_ingested = GandalfCleanIngestedSpectrum(gandalf_model_spectrum=model)
        gandalf = Gandalf(l1_spectra=l1spectra, model=specs.combined_model,
                          clean_model=clean_model, clean_observed=clean_ingested,
                          emission_line_model=emission,
                          tables=row, tables_replace=replacements)
        return gandalf, GandalfSpecs(model=model, ingested=ingested, emission=emission, clean_model=clean_model,
                                     clean_ingested=clean_ingested, nrow=specs.nrow)

    @classmethod
    def read_l2product_table(cls, this_fname, spectrum_hdu, row: CypherVariable, nrow,
                             parent_l1filenames,
                             IngestedSpectrumClass: Optional[Type[IngestedSpectrum]],
                             CombinedIngestedSpectrumClass: Optional[Type[CombinedIngestedSpectrum]],
                             ModelSpectrumClass: Optional[Type[ModelSpectrum]],
                             CombinedModelSpectrumClass: Optional[Type[BaseCombinedModelSpectrum]],
                             uses_disjoint_spectra: bool,
                             uses_combined_spectrum: Optional[bool],
                             formatter: str):
        # if the joint spectrum is not available, we dont read it, obvs
        if uses_combined_spectrum is None:
            uses_combined_spectrum = any('_C' in i for i in spectrum_hdu.data.names)
        fibre = Fibre.find(id=row['aps_id'])
        with unwind(CypherData(parent_l1filenames)) as l1_fname:  # for each parent l1 file
            l1file = L1File.find(fname=l1_fname)
            _, l1spectrum, fibretarget, _ = find_branch(l1file, L1Spectrum, FibreTarget, fibre)  # get the fibre target and l1 spectrum for this fibre and file
            colour_code = ArmConfig.find(anonymous_children=[l1spectrum])['colour_code']
            if uses_disjoint_spectra:
                individual = IngestedSpectrumClass(l1_spectrum=l1spectrum)
                individual_model = ModelSpectrumClass(uncombined_ingested_spectrum=individual)
        # now collect spec and models relative to the fibretarget
        if uses_disjoint_spectra:
            l1files, l1spectra, fibretargets, colour_codes, individuals, individual_models = collect(l1file, l1spectrum, fibretarget,
                                                                                    colour_code, individual, individual_model)
        else:
            l1files, l1spectra, fibretargets, colour_codes = collect(l1file, l1spectrum, fibretarget,  colour_code)
            individuals, individual_models = None, None
        if uses_combined_spectrum:
            combined = CombinedIngestedSpectrumClass(l1_spectra=[l1spectra[i] for i in range(len(parent_l1filenames))])
            combined_model = CombinedModelSpectrumClass(combined_ingested_spectrum=combined)
        else:
            combined, combined_model = None, None
        return FitSpecs(individuals, individual_models, combined, combined_model, colour_codes, nrow), l1spectra, fibretargets

    @classmethod
    def read_redrock(cls, this_fname, spectrum_hdu, colnames, safe_table: CypherVariable, parent_l1filenames, zs, **hiers):
        if len(spectrum_hdu.data) == 0:
            return
        replacements = column_name_acronym_replacements(colnames, 'rr')
        with unwind(safe_table, enumerated=True) as (row, nrow):
            specs, l1spectra, fibretargets = cls.read_l2product_table(this_fname, spectrum_hdu, row, nrow,
                                                          parent_l1filenames, UncombinedIngestedSpectrum,
                                                          CombinedIngestedSpectrum,
                                                          UncombinedModelSpectrum, CombinedModelSpectrum,
                                                          True, None, 'RR')
            redrock = cls.make_redrock_fit(l1spectra, specs, zs, row, len(parent_l1filenames), replacements)
            l2 = cls.make_l2(l1spectra, nspec=len(parent_l1filenames), fibre_target=fibretargets[0], **hiers)
            l2.attach_optionals(redrock=redrock)
        l2, redrocks, *r = collect(l2, redrock, *specs)
        return l2, redrocks, FitSpecs(*r), UncombinedIngestedSpectrum, CombinedIngestedSpectrum, UncombinedModelSpectrum, CombinedModelSpectrum

    @classmethod
    def read_rvspecfit(cls, this_fname, spectrum_hdu, colnames, safe_table: CypherVariable, parent_l1filenames, **hiers):
        if len(spectrum_hdu.data) == 0:
            return
        replacements = column_name_acronym_replacements(colnames, 'rvs')
        with unwind(safe_table, enumerated=True) as (row, nrow):
            rvs_specs, l1spectra, fibretargets = cls.read_l2product_table(this_fname, spectrum_hdu, row, nrow,
                                                      parent_l1filenames, UncombinedIngestedSpectrum,
                                                      CombinedIngestedSpectrum,
                                                      UncombinedModelSpectrum, CombinedModelSpectrum,
                                                      True, None, 'RVS')
            rvspecfit = RVSpecfit(l1_spectra=l1spectra, uncombined_model_spectra=[rvs_specs.individual_models[i] for i in range(len(parent_l1filenames))],
                                  combined_model_spectrum=rvs_specs.combined_model, tables=row, tables_replace=replacements)
            l2 = cls.make_l2(l1spectra, nspec=len(parent_l1filenames), fibre_target=fibretargets[0], **hiers)
            l2.attach_optionals(rvspecfit=rvspecfit)
        l2, rvspecfits, *r = collect(l2, rvspecfit, *rvs_specs)
        return l2, rvspecfits, FitSpecs(*r), UncombinedIngestedSpectrum, CombinedIngestedSpectrum, UncombinedModelSpectrum, CombinedModelSpectrum

    @classmethod
    def read_ferre(cls, this_fname, spectrum_hdu, colnames, safe_table: CypherVariable, parent_l1filenames, **hiers):
        if len(spectrum_hdu.data) == 0:
            return
        replacements = column_name_acronym_replacements(colnames, 'fr')
        with unwind(safe_table, enumerated=True) as (row, nrow):
            ferre_specs, l1spectra, fibretargets = cls.read_l2product_table(this_fname, spectrum_hdu, row, nrow,
                                                        parent_l1filenames, UncombinedIngestedSpectrum,
                                                        CombinedIngestedSpectrum,
                                                        UncombinedModelSpectrum, CombinedModelSpectrum,
                                                        True, None, 'FR')
            ferre = Ferre(l1_spectra=l1spectra, uncombined_model_spectra=[ferre_specs.individual_models[i] for i in range(len(parent_l1filenames))],
                          combined_model_spectrum=ferre_specs.combined_model, tables=row, tables_replace=replacements)
            l2 = cls.make_l2(l1spectra, nspec=len(parent_l1filenames), fibre_target=fibretargets[0], **hiers)
            l2.attach_optionals(ferre=ferre)
        l2, ferres, *r = collect(l2, ferre, *ferre_specs)
        return l2, ferres, FitSpecs(*r), UncombinedIngestedSpectrum, CombinedIngestedSpectrum, UncombinedModelSpectrum, CombinedModelSpectrum

    @classmethod
    def read_ppxf(cls, this_fname, spectrum_hdu, colnames, safe_table: CypherVariable, parent_l1filenames, **hiers):
        if len(spectrum_hdu.data) == 0:
            return
        replacements = column_name_acronym_replacements(colnames, 'ppxf')
        with unwind(safe_table, enumerated=True) as (row, nrow):
            ppxf_specs, l1spectra, fibretargets = cls.read_l2product_table(this_fname, spectrum_hdu, row, nrow,
                                                         parent_l1filenames, None,
                                                         CombinedIngestedSpectrum,
                                                         None, CombinedModelSpectrum,
                                                         False, True, 'PPXF')
            ppxf = PPXF(l1_spectra=l1spectra, combined_model_spectrum=ppxf_specs.combined_model, tables=row, tables_replace=replacements)
            l2 = cls.make_l2(l1spectra, nspec=len(parent_l1filenames), fibre_target=fibretargets[0], **hiers)
            l2.attach_optionals(ppxf=ppxf)
        l2, ppxfs, *r = collect(l2, ppxf, *ppxf_specs)
        return l2, ppxfs, FitSpecs(*r), None, CombinedIngestedSpectrum, None, CombinedModelSpectrum

    @classmethod
    def read_gandalf(cls, this_fname, spectrum_hdu, colnames, safe_table: CypherVariable, parent_l1filenames, **hiers):
        if len(spectrum_hdu.data) == 0:
            return
        replacements = column_name_acronym_replacements(colnames, 'gand')
        replacements.update(extract_lines_names(gandalf_line_names, colnames))
        replacements.update(extract_indices_names(gandalf_index_names, colnames))
        with unwind(safe_table, enumerated=True) as (row, nrow):
            gandalf_specs, l1spectra, fibretargets = cls.read_l2product_table(this_fname, spectrum_hdu, row, nrow,
                                                                              parent_l1filenames, None,
                                                                              CombinedIngestedSpectrum,
                                                                              None, GandalfModelSpectrum,
                                                                              False, True, 'GAND')
            gandalf, gandalf_extra_specs = cls.make_gandalf_structure(l1spectra, gandalf_specs, row, this_fname, nrow, replacements)
            l2 = cls.make_l2(l1spectra, nspec=len(parent_l1filenames), fibre_target=fibretargets[0], **hiers)
            l2.attach_optionals(gandalf=gandalf)
        l2, gandalfs, *r = collect(l2, gandalf, *gandalf_specs, *gandalf_extra_specs)
        return l2, gandalfs, FitSpecs(*r[:len(gandalf_specs)]), GandalfSpecs(*r[len(gandalf_specs):]), None, CombinedIngestedSpectrum, None, GandalfModelSpectrum

    @classmethod
    def make_redshift_arrays(cls, tbl: Table):
        templates = {col[4:] for col in tbl.colnames if 'czz_' in col and 'chi2' not in col}
        row = tbl[0]
        d = {}
        for template_name in templates:
            zs = row[f'czz_{template_name}']
            array = RedshiftArray(value=CypherData(zs), start=zs[0], end=zs[1], step=zs[1] - zs[0])
            d[template_name] = array
        return d

    @classmethod
    def read(cls, directory: Union[Path, str], fname: Union[Path, str], slc: slice = None, part=None):
        if part is None:
            raise RuntimeError(f"{cls.__name__}.read() requires a part argument otherwise py2neo will crash")
        if part not in cls.parts:
            raise ValueError(f"Unrecognised part {part}. Only {cls.parts} are allowed")
        fname = Path(fname)
        directory = Path(directory)
        path = directory / fname
        header, aps = cls.read_header_and_aps(path)
        l1files = cls.parse_fname(header, fname)
        aps = APS(version=aps)
        hierarchies = cls.find_shared_hierarchy(path)
        astropy_hdus = fits.open(path)
        fnames = [l1.fname for l1 in l1files]
        assert len(fnames) > 1, f"{fname} has only one L1 file"
        safe_tables = {}
        safe_cypher_tables = {}
        for i, hdu in enumerate(astropy_hdus[1:4], 1):
            safe_tables[i] = filter_products_from_table(Table(hdu.data)[slc], MAX_REDSHIFT_GRID_LENGTH)
            cols = [col for col in safe_tables[i].colnames if not ('chi2' not in col and 'czz_' in col)]
            safe_cypher_tables[i] =  CypherData(safe_tables[i][cols])
        if part == 'RVS':
            l2, specfits, specs, *types = cls.read_rvspecfit(path, astropy_hdus[5], [x.lower() for x in astropy_hdus[2].data.names],
                                                 safe_cypher_tables[2], fnames, **hierarchies)
            hdu_node = 5
        elif part == 'FR':
            l2, specfits, specs, *types = cls.read_ferre(path, astropy_hdus[5], [x.lower() for x in astropy_hdus[2].data.names],
                                             safe_cypher_tables[2], fnames, **hierarchies)
            hdu_node = 5
        elif part == 'PPXF':
            l2, specfits, specs, *types = cls.read_ppxf(path, astropy_hdus[6], [x.lower() for x in astropy_hdus[3].data.names],
                                            safe_cypher_tables[3], fnames, **hierarchies)
            hdu_node = 6
        elif part == 'RR':
            zs = cls.make_redshift_arrays(safe_tables[1])
            l2, specfits, specs, *types = cls.read_redrock(path, astropy_hdus[4], [x.lower() for x in astropy_hdus[1].data.names],
                                               safe_cypher_tables[1], fnames, zs, **hierarchies)
            hdu_node = 4
        elif part == 'GAND':
            l2, specfits, specs, extra_specs, *types = cls.read_gandalf(path, astropy_hdus[6], [x.lower() for x in astropy_hdus[3].data.names],
                             safe_cypher_tables[3], fnames, **hierarchies)
            hdu_node = 6
        else:
            raise ValueError(f"{part} is not a valid part")
        hdu_nodes, file, _ = cls.read_hdus(directory, fname, l2=l2, l1files=l1files, aps=aps, **hierarchies)
        hdu = hdu_nodes[hdu_node]
        names = {'logwvl': 'loglam', 'wvl': 'lambda'}
        if specs is not None:
            suffix = '' if part in ['PPXF', 'GAND'] else '_C'
            cls.attach_products_to_spectra(specs, part, hdu, names, types, suffix)
        if part == 'GAND':
            cls.attach_products_to_gandalf_extra_spectra(extra_specs, hdu)


class L2SingleFile(L2File):
    singular_name = 'l2single_file'
    children = [APS]
    parents = [Multiple(L1SingleFile, 2, 3, constrain=(Exposure,)), Multiple(L2Single, maxnumber=1000)]
    L2 = L2Single
    L1 = L1SingleSpectrum


    @classmethod
    def find_shared_hierarchy(cls, path) -> Dict:
        header = cls.read_header_and_aps(path)[0]
        runids = cls.parser_ftypes_runids(header)[1]
        run = Run.find(id=runids[0])
        return {'exposure': Exposure.find(anonymous_children=[run])}


class L2StackFile(L2File):
    singular_name = 'l2stack_file'
    children = [APS]
    parents = [Multiple(L1StackFile, 1, 3, constrain=(OB,)), Multiple(L2Stack, maxnumber=1000)]
    L2 = L2Stack
    L1 = L1StackSpectrum

    @classmethod
    def find_shared_hierarchy(cls, path) -> Dict:
        header = cls.read_header_and_aps(path)[0]
        return {'ob': OB.find(id=int(header['OBID']))}


class L2SuperstackFile(L2File):
    singular_name = 'l2superstack_file'
    children = [APS]
    parents = [Multiple(L1SingleFile, 0, 3, constrain=(OBSpec,)),
               Multiple(L1StackFile, 0, 3, constrain=(OBSpec,)),
               Multiple(L1SuperstackFile, 0, 3, constrain=(OBSpec,)),
               Multiple(L2Superstack, maxnumber=1000)]
    L2 = L2Superstack
    L1 = L1Spectrum

    @classmethod
    def find_shared_hierarchy(cls, path) -> Dict:
        header = cls.read_header_and_aps(path)[0]
        return {'obspec': OBSpec.find(xml=str(header['cat-name']))}


class L2SupertargetFile(L2File):
    singular_name = 'l2supertarget_file'
    match_pattern = 'WVE_*aps.fits'
    children = [APS]
    parents = [Multiple(L1SupertargetFile, 2, 3, constrain=(WeaveTarget,)), L2Supertarget]
    L2 = L2Supertarget
    L1 = L1SupertargetSpectrum

    @classmethod
    def parse_fname(cls, header, fname, instantiate=True) -> List[L1File]:
        raise NotImplementedError

    @classmethod
    def find_shared_hierarchy(cls, path: Path) -> Dict:
        hdus = fits.open(path)
        names = [i.name for i in hdus]
        cname = hdus[names.index('CLASS_TABLE')].data['CNAME'][0]
        return {'weavetarget': WeaveTarget.find(cname=cname)}


hierarchies = [i[-1] for i in inspect.getmembers(sys.modules[__name__], _predicate)]


import inspect
import sys

from weaveio.hierarchy import Optional, Multiple, Hierarchy, OneOf
from weaveio.opr3.hierarchy import Spectrum, Single, FibreTarget, Stacked, \
    Stack, OB, ArmConfig, Superstack, OBSpec, Supertarget, WeaveTarget, RawSpectrum, _predicate, Spectrum1D, MeanFlux


class L1(Hierarchy):
    is_template = True


class L1Spectrum(Spectrum1D, L1):
    is_template = True
    children = Spectrum1D.children + [Optional('self', idname='adjunct', one2one=True)]
    products = ['flux', 'ivar', 'sensfunc']
    factors = Spectrum.factors + ['nspec', 'snr'] + MeanFlux.as_factors('g', 'r', 'i', 'gg', 'bp', 'rp', prefix='mean_flux_')
    parents = [FibreTarget]
    indexes = ['nspec']


class NoSS(Spectrum1D):
    plural_name = 'nosses'
    singular_name = 'noss'
    products = ['flux', 'ivar']
    parents = [OneOf(L1Spectrum, one2one=True)]
    children = [Optional('self', idname='adjunct', one2one=True)]
    identifier_builder = ['l1_spectrum']
    indexes = [None]  # removes nspec index


class L1SingleSpectrum(L1Spectrum, Single):
    """
    A single spectrum row processed from a raw spectrum, belonging to one fibretarget and one run.
    """
    singular_name = 'l1single_spectrum'
    plural_name = 'l1single_spectra'
    parents = [FibreTarget, RawSpectrum, ArmConfig] + L1Spectrum.parents  # order matters for speed! filters by fibretarget, then rawspectrum, then armconfig
    identifier_builder = ['raw_spectrum', 'fibre_target', 'arm_config']
    factors = L1Spectrum.factors + [
        'rms_arc1', 'rms_arc2', 'resol', 'helio_cor',
        'wave_cor1', 'wave_corrms1', 'wave_cor2', 'wave_corrms2',
        'skyline_off1', 'skyline_rms1', 'skyline_off2', 'skyline_rms2',
        'sky_shift', 'sky_scale']


class L1StackedSpectrum(L1Spectrum, Stacked):
    is_template = True
    singular_name = 'l1stacked_spectrum'
    plural_name = 'l1stacked_spectra'
    parents = [Multiple(L1SingleSpectrum, 2, 10, constrain=(FibreTarget, ArmConfig, ))] + L1Spectrum.parents
    identifier_builder = ['l1single_spectra']


class L1StackSpectrum(L1StackedSpectrum, Stack):
    """
    A stacked spectrum row processed from > 1 single spectrum, belonging to one fibretarget but many runs within the same OB.
    """
    singular_name = 'l1stack_spectrum'
    plural_name = 'l1stack_spectra'
    parents = [Multiple(L1SingleSpectrum, 2, 10, constrain=(FibreTarget, OB, ArmConfig))] + L1Spectrum.parents


class L1SuperstackSpectrum(L1StackedSpectrum, Superstack):
    """
    A stacked spectrum row processed from > 1 single spectrum, belonging to one fibretarget but many runs within the same OBSpec.
    """
    singular_name = 'l1superstack_spectrum'
    plural_name = 'l1superstack_spectra'
    parents = [Multiple(L1SingleSpectrum, 2, 10, constrain=(FibreTarget, OBSpec, ArmConfig))] + L1Spectrum.parents


class L1SupertargetSpectrum(L1StackedSpectrum, Supertarget):
    """
    A stacked spectrum row processed from > 1 single spectrum, belonging to one weavetarget over many different OBSpecs.
    A weavetarget can have more than supertarget spectrum so an ID is added to the filename
    todo: This structure needs looking at.
    """
    singular_name = 'l1supertarget_spectrum'
    plural_name = 'l1supertarget_spectra'
    factors = ['id']
    parents = [Multiple(L1SingleSpectrum, 2, 10, constrain=(WeaveTarget, ArmConfig))] + L1Spectrum.parents
    identifier_builder = ['l1single_spectra', 'id']
    indexes = []



hierarchies = [i[-1] for i in inspect.getmembers(sys.modules[__name__], _predicate)]


import logging
import os
import re
import time
from collections import defaultdict
from contextlib import contextmanager
from functools import reduce
from pathlib import Path
from textwrap import dedent
from typing import Union, List, Tuple, Type, Dict, Set, Callable, Optional
from uuid import uuid4

import graphviz
import networkx as nx
import pandas as pd
import py2neo
import textdistance
from networkx import NetworkXNoPath, NodeNotFound
from networkx.drawing.nx_pydot import to_pydot
from operator import mul
from py2neo import ClientError, DatabaseError
from tqdm import tqdm

from .file import File, HDU
from .graph import Graph, _convert_datatypes
from .hierarchy import Multiple, Hierarchy, Graphable, OneOf
from .path_finding import HierarchyGraph, get_all_class_bases
from .readquery import Query
from .readquery.exceptions import UserError, CardinalityError
from .readquery.results import RowParser
from .utilities import make_plural, make_singular
from .writequery import Unwind

CONSTRAINT_FAILURE = re.compile(r"already exists with label `(?P<label>[^`]+)` and property "
                                r"`(?P<idname>[^`]+)` = (?P<idvalue>[^`]+)$", flags=re.IGNORECASE)


class ConnectionTimeOutError(ConnectionError):
    pass


def process_neo4j_error(data: 'Data', file: File, msg):
    matches = CONSTRAINT_FAILURE.findall(msg)
    if not len(matches):
        return  # cannot help
    label, idname, idvalue = matches[0]
    # get the node properties that already exist
    extant = data.graph.neograph.evaluate(f'MATCH (n:{label} {{{idname}: {idvalue}}}) RETURN properties(n)')
    fname = data.graph.neograph.evaluate(f'MATCH (n:{label} {{{idname}: {idvalue}}})-[*]->(f:File) return f.fname limit 1')
    idvalue = idvalue.strip("'").strip('"')
    file.data = data
    obj = [i for i in data.hierarchies if i.__name__ == label][0]
    instance_list = getattr(file, obj.plural_name)
    new = {}
    if not isinstance(instance_list, (list, tuple)):  # has an unwind table object
        new_idvalue = instance_list.identifier
        if isinstance(new_idvalue, Unwind):
            # find the index in the table and get the properties
            filt = (new_idvalue.data == idvalue).iloc[:, 0]
            for k in extant.keys():
                if k == 'id':
                    k = idname
                value = getattr(instance_list, k, None)
                if isinstance(value, Unwind):
                    table = value.data.where(pd.notnull(value.data), 'NaN')
                    new[k] = str(table[k][filt].values[0])
                else:
                    new[k] = str(value)
        else:
            # if the identifier of this object is not looping through a table, we cant proceed
            return
    else:  # is a list of non-table things
        found = [i for i in instance_list if i.identifier == idvalue][0]
        for k in extant.keys():
            value = getattr(found, k, None)
            new[k] = value
    comparison = pd.concat([pd.Series(extant, name='extant'), pd.Series(new, name='to_add')], axis=1)
    filt = comparison.extant != comparison.to_add
    filt &= ~comparison.isnull().all(axis=1)
    where_different = comparison[filt]
    logging.exception(f"The node (:{label} {{{idname}: {idvalue}}}) tried to be created twice with different properties.")
    logging.exception(f"{where_different}")
    logging.exception(f"filenames: {fname}, {file.fname}")


def get_all_subclasses(cls: Type[Graphable]) -> List[Type[Graphable]]:
    all_subclasses = []
    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))
    return all_subclasses


def find_children_of(parent):
    hierarchies = get_all_subclasses(Hierarchy)
    children = set()
    for h in hierarchies:
        if len(h.parents):
            if any(p is parent if isinstance(p, type) else p.node is parent for p in h.parents):
                children.add(h)
    return children


class IndirectAccessError(Exception):
    pass


class MultiplicityError(Exception):
    pass


def is_multiple_edge(graph, x, y):
    return not graph.edges[(x, y)]['multiplicity']

def expand_template_relation(relation, allowed_hiers: Set[Type[Hierarchy]]):
    """
    Returns a list of relations that relate to each non-template class
    e.g.
    >>> expand_template_relation(Multiple(L1StackedSpectrum))
    [Multiple(L1SingleSpectrum), Multiple(L1StackSpectrum), Multiple(L1SuperstackSpectrum)]
    """
    if not relation.node.is_template:
        return [relation]
    subclasses = [cls for cls in get_all_subclasses(relation.node) if not cls.is_template and cls in allowed_hiers]
    return [Multiple(subclass, 0, relation.maxnumber, relation.constrain, relation.relation_idname, relation.one2one) for subclass in subclasses]


def add_relation_graph_edge(graph, parent, child, relation: Multiple, allowed_hiers: Set[Type[Hierarchy]]):
    """
    if an object of type O requires n parents of type P then this is equivalent to defining that instances of those behave as:
        P-(n)->O (1 object of type O has n parents of type P)
    it implicitly follows that:
        O--(m)--P (each of object's parents of type P can be used by an unknown number `m` of objects of type O = many to one)
    if an object of type O requires n children of type C then this is equivalent to defining that instances of those behave as:
        O-(n)->C (1 object has n children of type C)
        it implicitly follows that:
            child-[m]->Object (each child has m parents of type O)
    """
    relation.instantate_node()
    child_defines_parents = relation.node is parent
    for relation in expand_template_relation(relation, allowed_hiers):
        relation.instantate_node()
        # only parent-->child is in the database
        relstyle = 'solid' if relation.maxnumber == 1 else 'dashed'
        if parent is child:
            for a, b in [(parent, child), (child, parent)]:
                graph.add_edge(a, b, singular=relation.maxnumber == 1,
                               optional=relation.minnumber == 0, style=relstyle, actual_number=1)
        elif child_defines_parents:  # i.e. parents = [...] is set in the class for this object
            # child instance has n of type Parent, parent instance has unknown number of type Child
            parent = relation.node  # reset from new relations
            graph.add_edge(child, parent, singular=relation.maxnumber == 1,
                           optional=relation.minnumber == 0, style=relstyle, actual_number=1)
            if relation.one2one:
                graph.add_edge(parent, child, singular=True, optional=True, style='solid',
                               relation=relation, actual_number=1)
            else:
                graph.add_edge(parent, child, singular=False, optional=True, style='dotted',
                               relation=relation, actual_number=1)
        else:  # i.e. children = [...] is set in the class for this object
            # parent instance has n of type Child, each child instance has one of type Parent
            child = relation.node  # reset from new relations
            graph.add_edge(parent, child, singular=relation.maxnumber == 1,
                           optional=relation.minnumber == 0,
                           relation=relation, style=relstyle, actual_number=1)
            graph.add_edge(child, parent, singular=True, optional=True, style='solid', actual_number=1)


def make_relation_graph(hierarchies: Set[Type[Hierarchy]]):
    graph = nx.DiGraph()
    for h in hierarchies:
        if h not in graph.nodes:
            graph.add_node(h)
        for child in h.children:
            rel = child if isinstance(child, Multiple) else OneOf(child)
            child = child.node if isinstance(child, Multiple) else child
            add_relation_graph_edge(graph, h, child, rel, hierarchies)
        for parent in h.parents:
            rel = parent if isinstance(parent, Multiple) else OneOf(parent)
            parent = parent.node if isinstance(parent, Multiple) else parent
            add_relation_graph_edge(graph, parent, h, rel, hierarchies)
    return graph

def hierarchies_from_hierarchy(hier: Type[Hierarchy], done=None, templates=False) -> Set[Type[Hierarchy]]:
    if done is None:
        done = []
    hierarchies = set()
    todo = {h.node if isinstance(h, Multiple) else h for h in hier.parents + hier.children + hier.produces}
    if not templates:
        todo = {h for h in todo if not h.is_template}
    else:
        todo.update({h for h in todo for hh in get_all_class_bases(h) if issubclass(hh, Hierarchy)})
    for new in todo:
        if isinstance(new, Multiple):
            new.instantate_node()
            h = new.node
        else:
            h = new
        if h not in done and h is not hier:
            hierarchies.update(hierarchies_from_hierarchy(h, done, templates))
            done.append(h)
    hierarchies.add(hier)
    return hierarchies

def hierarchies_from_files(*files: Type[File], templates=False) -> Set[Type[Hierarchy]]:
    hiers = {h.node if isinstance(h, Multiple) else h for file in files for h in file.children + file.produces}
    if not templates:
        hiers = {h for h in hiers if not h.is_template}
    else:
        hiers.update({h for h in hiers for hh in get_all_class_bases(h) if issubclass(hh, Hierarchy)})
    hiers.update(set(files))
    return reduce(set.union, map(hierarchies_from_hierarchy, hiers))

def make_arrows(path, forwards: List[bool], descriptors=None):
    descriptors = [descriptors]*len(forwards) if not isinstance(descriptors, (list, tuple)) else descriptors
    descriptors = [f":{descriptor}" if descriptor is not None else "" for descriptor in descriptors]
    assert len(forwards) == len(path) - 1
    forward_arrow = '-[{name}{descriptor}]->'
    backward_arrow = '<-[{name}{descriptor}]-'
    nodes = list(map('(:{})'.format, [p.__name__ for p in path[1:]]))
    path_list = ['({in_var})']
    for i, (node, forward) in enumerate(zip(nodes, forwards)):
        arrow = forward_arrow if forward else backward_arrow
        if i == len(forwards) - 1:
            arrow = arrow.format(name='{name}', descriptor=descriptors[i])
        else:
            arrow = arrow.format(name="", descriptor=descriptors[i])
        path_list.append(arrow)
        path_list.append(node)
    path_list[-1] = path_list[-1].replace('(:', '({out_var}:')
    return ''.join(path_list)


def plot_graph(G, fname, format):
    return graphviz.Source(to_pydot(G).to_string()).render(fname, format=format)


class Data:
    filetypes = []
    extra_hierarchies = []

    def __new__(cls, *args, **kwargs):
        from fastapi.encoders import ENCODERS_BY_TYPE
        ENCODERS_BY_TYPE[cls] = lambda x: x.as_dict()
        return super().__new__(cls)

    def __init__(self, rootdir: Union[Path, str] = None,
                 host: str = None, port=None, dbname=None,
                 password=None, user=None, verbose=False):
        self.verbose = verbose
        self.dbname = dbname or os.getenv('WEAVEIO_DB', 'production')
        self.host = host or os.getenv('WEAVEIO_HOST', '127.0.0.1')
        self.port = port or os.getenv('WEAVEIO_PORT', 7687)
        self.password = password or os.getenv('WEAVEIO_PASSWORD', 'weavepassword')
        self.user = user or os.getenv('WEAVEIO_USER', 'weaveuser')
        rootdir = rootdir or os.getenv('WEAVEIO_ROOTDIR')
        if rootdir is None:
            raise ValueError(f"You must specify WEAVEIO_ROOTDIR as an environment variable or as an argument to Data (rootdir=...)")
        self.rootdir = Path(rootdir)
        self.write_allowed = False
        self.query = Query(self)
        self.rowparser = RowParser(self.rootdir)
        self.filelists = {}
        self.hierarchy_graph = HierarchyGraph()
        self.hierarchy_graph.initialise()
        if self.filetypes:
            self.hierarchies = hierarchies_from_files(*self.filetypes, templates=True)
        else:
            self.hierarchies = set()
        self.hierarchies.update(set(self.extra_hierarchies))
        self.hierarchies.update({hh for h in self.hierarchies for hh in get_all_class_bases(h)})
        self.class_hierarchies = {h.__name__: h for h in self.hierarchies}
        self.singular_hierarchies = {h.singular_name: h for h in self.hierarchies}  # type: Dict[str, Type[Hierarchy]]
        self.plural_hierarchies = {h.plural_name: h for h in self.hierarchies if h.plural_name != 'graphables'}
        self.factor_hierarchies = defaultdict(set)
        for h in self.hierarchies:
            for f in getattr(h, 'products_and_factors', []):
                self.factor_hierarchies[f.lower()].add(h)
            if h.idname is not None:
                self.factor_hierarchies[h.idname].add(h)
        self.factor_hierarchies = dict(self.factor_hierarchies)  # make sure we always get keyerrors when necessary!
        self.factors = set(self.factor_hierarchies.keys())
        self.plural_factors =  {make_plural(f.lower()): f.lower() for f in self.factors}
        self.singular_factors = {f.lower() : f.lower() for f in self.factors}
        self.singular_idnames = {h.idname: h for h in self.hierarchies if h.idname is not None}
        self.plural_idnames = {make_plural(k): v for k,v in self.singular_idnames.items()}
        self.relative_names = defaultdict(dict)
        for h in self.hierarchies:
            for name, relation in h.relative_names.items():
                self.relative_names[name][h.__name__] = relation
        self.relative_names = dict(self.relative_names)
        self.plural_relative_names = {make_plural(name): name for name in self.relative_names}

    def __repr__(self):
        return f'<Data({self.user}@{self.host}[{self.dbname}]:{self.port}:{self.rootdir})>'

    def as_dict(self):
        return {'user': self.user, 'host': self.host, 'dbname': self.dbname, 'port': self.port, 'rootdir': self.rootdir}

    def __dir__(self) -> List[str]:
        neighbors = [i.plural_name for i in self.hierarchies]
        neighbors += [i.lower() for h in self.hierarchies for i in h.products_and_factors]
        neighbors = [i if '[' not in i else f'"{i}"' for i in neighbors]
        return neighbors

    def _ipython_key_completions_(self):
        return [i.strip('"') for i in self.__dir__()]

    @property
    def verbose(self):
        return self._verbose

    @verbose.setter
    def verbose(self, value):
        self._verbose = value
        if value:
            logging.basicConfig(level=logging.INFO)

    # noinspection PyTypeHints
    def expand_template_object(self, obj: str) -> Set[str]:
        obj = self.singular_hierarchies[self.singular_name(obj)]
        return {h.__name__ for h in self.hierarchies if issubclass(h, obj) and not h.is_template}

    def expand_dependent_object(self, obj: str, *dependencies: str) -> Set[str]:
        obj = self.singular_hierarchies[self.singular_name(obj)]
        dependencies = {self.singular_hierarchies[self.singular_name(o)] for o in dependencies}
        parents = self.parents_of_defined_child(obj)
        expanded_dependencies = set()
        for o in dependencies:
            if o.is_template:
                expanded_dependencies |= set(get_all_subclasses(o))
            else:
                expanded_dependencies.add(o)
        return {o.__name__ for o in parents if o in expanded_dependencies}

    def _path_to_hierarchy(self, from_obj: Type[Hierarchy], to_obj:  Type[Hierarchy], singular: bool):
        """
        When searching for a path, the target is either above or below the source in one direction only
        If the target is defined as a child by another, then the search is redefined as for the predecessors of that child
            i.e. ob.nosses => [ob.l1single_spectra.noss, ob.l1stack_spectra.noss, ob.l1superstack_spectra.noss, ...]
                 this is then an error since it is an ambiguous path
        If the source is defined as a child by another, then additionally, the other node is required to define a path
            i.e. ...noss.obs requires that ... is known.
                 this is therefore an error since it is an ambiguous path
        If force_single, use only single edges
        for {singles, multiples}:
            Find the shortest path in one direction, but search both directions, both are equally valid
            If more than one path is returned, throw an ambiguous path exception

        """
        paths = list(self.hierarchy_graph.find_paths(from_obj, to_obj, singular))
        if not paths:
            if singular:
                raise NetworkXNoPath(f"No singular path found between `{from_obj}` and `{to_obj}`")
            raise NetworkXNoPath(f"No path found between `{from_obj}` and `{to_obj}`")
        paths, edges, reversed = zip(*[(path[::-1], [self.hierarchy_graph.short_edge(e[1], e[0]) for e in es[::-1]], True) if path[0] is to_obj else (path, es, False) for path, es in paths])
        singulars = [self.hierarchy_graph.edge_path_is_singular(es) for es in edges]
        return paths, singulars, reversed


    def parents_of_defined_child(self, potential_child: Type[Hierarchy]) -> Set[Type[Hierarchy]]:
        parents = {h for h in self.hierarchies if potential_child in [c.node if isinstance(c, Multiple) else c for c in h.children]}
        return {p for p in parents if not issubclass(p, File) and p is not potential_child}

    def is_generic_object(self, obj: str) -> bool:
        try:
            h = self.singular_hierarchies[self.singular_name(obj)]
            return h.is_template or self.parents_of_defined_child(h)
        except KeyError:
            return False

    def paths_to_hierarchy(self, from_obj: str, to_obj: str, singular: bool, descriptor=None, return_objs=False):
        a, b = map(self.singular_name, [from_obj, to_obj])
        from_obj, to_obj = self.singular_hierarchies[a], self.singular_hierarchies[b]
        try:
            paths, singulars, reversed = self._path_to_hierarchy(from_obj, to_obj, singular)
            arrows = [make_arrows(path, [not r]*(len(path)-1), descriptor) for path, r in zip(paths, reversed)]
            if return_objs:
                return arrows, singulars, paths
            return arrows, singulars
        except nx.NetworkXNoPath:
            if not singular:
                to = f"multiple `{self.plural_name(b)}`"
            else:
                to = f"only one `{self.singular_name(b)}`"
            from_ = self.singular_name(a.lower())
            raise NetworkXNoPath(f"Can't find a link between `{from_}` and {to}. "
                                f"This may be because it doesn't make sense for `{from_}` to have {to}. "
                                f"Try checking the cardinality of your query.")

    def all_links_to_hierarchy(self, hierarchy: Type[Hierarchy], edge_constraint: Callable[[nx.DiGraph, Tuple], bool]) -> Set[Type[Hierarchy]]:
        hierarchy = self.class_hierarchies[self.class_name(hierarchy)]
        g = self.relation_graphs[-1]
        singles = nx.subgraph_view(g, filter_edge=lambda a, b: edge_constraint(g, (a, b)))
        hiers = set()
        for node in singles.nodes:
            if nx.has_path(singles, hierarchy, node):
                hiers.add(node)
        return hiers

    def all_single_links_to_hierarchy(self, hierarchy: Type[Hierarchy]) -> Set[Type[Hierarchy]]:
        return self.all_links_to_hierarchy(hierarchy, lambda g, e: g.edges[e]['singular'])

    def all_multiple_links_to_hierarchy(self, hierarchy: Type[Hierarchy]) -> Set[Type[Hierarchy]]:
        return self.all_links_to_hierarchy(hierarchy, lambda g, e: not g.edges[e]['singular'])

    def write_cypher(self, collision_manager='track&flag'):
        if self.write_allowed:
            return self.graph.write(collision_manager)
        raise IOError(f"You have not allowed write operations in this instance of data (write=False)")

    @property
    @contextmanager
    def write(self):
        self.write_allowed = True
        yield self.graph
        self.write_allowed = False

    def is_unique_factor(self, name):
        return len(self.factor_hierarchies[name]) == 1

    @property
    def graph(self):
        d = {}
        if self.password is not None:
            d['password'] = self.password
        if self.user is not None:
            d['user'] = self.user
        return Graph(host=self.host, port=self.port, name=self.dbname, write=self.write_allowed, **d)

    def make_constraints_cypher(self) -> Dict[str, List[str]]:
        d = {hierarchy: hierarchy.make_schema() for hierarchy in self.hierarchies}
        d['temporary'] = ['CREATE CONSTRAINT ON (t:TemporaryMerge) ASSERT t.id IS UNIQUE']
        d['dbcreated'] = ['CREATE INDEX n_dbcreated FOR (n:Hierarchy) ON n.`_dbcreated`']
        d['dbupdated'] = ['CREATE INDEX n_dbupdated FOR (n:Hierarchy) ON n.`_dbupdated`']
        return d

    def apply_constraints(self, silent=False):
        if not self.write_allowed:
            raise IOError(f"Writing is not allowed")
        templates = set()
        equivalencies = set()
        for hier, qs in tqdm(self.make_constraints_cypher().items(), desc='applying constraints', disable=silent):
            if not qs:
                templates.add(hier)
            for q in qs:
                try:
                    self.graph.neograph.run(q)
                except py2neo.ClientError as e:
                    if '[Schema.EquivalentSchemaRuleAlreadyExists]' in str(e):
                        equivalencies.add(hier)
                        templates.add(hier)
        total = len(templates)
        templates -= equivalencies
        if len(equivalencies):
            n = f'({len(equivalencies)}/{total})'
            equivalencies = ', '.join(sorted([getattr(t, '__name__', str(t)) for t in equivalencies]))
            logging.info(f'EquivalentSchemaRuleAlreadyExists for {n}: {equivalencies}')
        if len(templates):
            n = f'({len(templates)}/{total})'
            templates = ', '.join(sorted([getattr(t, '__name__', str(t)) for t in templates]))
            logging.info(f'No index/constraint exists for {n}: {templates}')
        self.await_indexes()

    def await_indexes(self, timeout=300):
        logging.info("Waiting for indexes to come online...")
        self.graph.neograph.run(f'call db.awaitIndexes({timeout})')
        logging.info("Indexes are online")

    def drop_all_constraints(self, indexes=True):
        if not self.write_allowed:
            raise IOError(f"Writing is not allowed")
        constraints = self.graph.neograph.run('CALL db.constraints() YIELD name return "DROP CONSTRAINT " + name + ";"')
        for constraint in tqdm(constraints, desc='dropping constraints'):
            self.graph.neograph.run(str(constraint)[1:-1])
        if indexes:
            self.drop_all_indexes()

    def drop_all_indexes(self):
        if not self.write_allowed:
            raise IOError(f"Writing is not allowed")
        indexes = self.graph.neograph.run('CALL db.indexes() yield name, labelsOrTypes where size(labelsOrTypes) > 0 return "DROP INDEX " + name + ";"')
        for index in tqdm(indexes, desc='dropping indexes'):
            self.graph.neograph.run(str(index)[1:-1])

    def get_complete_files(self, *filetypes, return_only_fname=True):
        filetypes = [self.singular_hierarchies[self.singular_name(f)].__name__ for f in filetypes]
        if not filetypes:
            filetypes = ['File']
        r = 'fname' if return_only_fname else 'path'
        q = f"MATCH (f:File) where any(l in labels(f) where l in $filetypes) and f._dbcompleted_complete RETURN DISTINCT f.{r}"
        result = self.graph.execute(q, filetypes=filetypes).to_series(0).values.tolist()
        if return_only_fname:
            return result
        return list(map(lambda x: self.rootdir / x, result))

    def raise_collisions(self):
        """
        returns the properties that would have been overwritten in nodes and relationships.
        """
        node_collisions = self.graph.execute("MATCH (c: _Collision) return c { .*}").to_data_frame()
        rel_collisions = self.graph.execute("MATCH ()-[c: _Collision]-() return c { .*}").to_data_frame()
        return node_collisions, rel_collisions

    def get_state_of_file(self, fname: str) -> int:
        """Returns the state (timestamp) of the creation of the last data product belong to a given file"""
        c = self.data.execute("match (f: File {fname: $fname}) optional match (f)<-[:is_required_by]-(h:Hierarchy) "
                          "where not h:HDU with  h.`_dbcreated` as d order by d desc return d limit 1", fname=fname)
        return c.evaluate()

    def restore_state(self, state: int):
        """
        Restores database to the given state (timestamp). Danger zone here.
        """
        if not self.write_allowed:
            raise IOError(f"Writing is not allowed")
        return self.graph.execute('match (n) where n.`_dbcreated` > $timestamp detach delete n', timestamp=state)

    def mark_batch_complete_query(self, fname, slice, part, total_length, all_parts):
        """
        Mark fname[slice][part] as complete in the db and if all parts are then complete, mark that too
        file._dbcompleted_{part} is a list of rows that have been marked as complete for that part
        file._dbcompleted_{part}_complete = True if that part is completely written
        file._dbcompleted_complete = True all parts are completely written
        part is "None" for files which have no parts
        """
        def get_intervals(v):
            return f"[c in apoc.coll.pairs([0] + [x in range(0, size({v})-2) where {v}[x] <> {v}[x+1] -1 | x+1]) | CASE WHEN c[1] is null THEN [{v}[c[0]],{v}[-1]] ELSE [{v}[c[0]], {v}[c[1]-1]] END]"
        def get_list(v):
            return f"apoc.coll.flatten([i in {v} | range(i[0],i[1])])"
        q = dedent(rf"""
        MATCH (f: File {{fname: $_check_fname}})
        WITH *, apoc.coll.zip(f.`_dbcompleted_{part}_start`, f.`_dbcompleted_{part}_end`) as already_intervals
        WITH *, {get_list('already_intervals')} as already_done
        WITH *, apoc.coll.sort(apoc.coll.toSet(coalesce(already_done, []) + $_check_completed))+[-1] as done
        WITH *, {get_intervals('done')} as done_intervals
        WITH *, CASE WHEN size(done_intervals) = 0 THEN [[0, $_check_total_length-1]] ELSE done_intervals[..-1] END as done_intervals
        SET f.`_dbcompleted_{part}_start` = [i in done_intervals | i[0]]
        SET f.`_dbcompleted_{part}_end` = [i in done_intervals | i[1]]
        SET f.`_dbcompleted_{part}_complete` = done_intervals = [[0, $_check_total_length-1]]
        SET f.`_dbcompleted_complete` = all(x in $_check_parts where f['_dbcompleted_'+x+'_complete'])
        """)
        completed = list(range(0, total_length)[slice])
        return q, {'_check_fname': fname, '_check_completed': completed,
                   '_check_parts': list(map(str, all_parts)), '_check_total_length': total_length}

    def file_batch_is_complete(self, filetype: File, path: Path, slc: slice, part: str) -> bool:
        """
        Returns True if the file[rows-slice][part] is complete otherwise False
        """
        total_length = filetype.length(self.rootdir / path, part)
        q = dedent(f"""
        OPTIONAL MATCH (f:File {{fname: $fname}})
        with f, apoc.coll.zip(f._dbcompleted_{part}_start, f._dbcompleted_{part}_end) as intervals
        return all(x in $rows where any(i in intervals where i[0] <= x and x <= i[1]))
        """)
        return self.graph.execute(q, fname=path.name, rows=list(range(total_length)[slc])).evaluate() or False

    def files_to_batches(self, *paths: Union[Path, str], batch_size=None, batches_slc=None, parts=None,
                         skip_complete=False):
        batches = []
        if not paths:
            return []
        if len(paths) == 1 and isinstance(paths[0], (tuple, list)):
            paths = paths[0]
        bar = tqdm(paths, desc='Building todo list')
        for path in bar:
            path = Path(path)
            matches = [f for f in self.filetypes if f.match_file(self.rootdir, path.relative_to(self.rootdir), self.graph)]
            if len(matches) > 1:
                raise ValueError(f"{path} matches more than 1 file type: {matches} with `{[m.match_pattern for m in matches]}`")
            filetype = matches[0]
            filetype_batch_size = filetype.recommended_batchsize if batch_size is None else batch_size
            slices = filetype.get_batches(path, filetype_batch_size, parts, batches_slc)
            bs = [(filetype, path.relative_to(self.rootdir), slc, part) for slc, part in slices]
            if skip_complete:
                bs = [b for b in bs if not self.file_batch_is_complete(*b)]
            batches += bs
        batches.sort(key=lambda b: (self.filetypes.index(b[0]), b[1]))  # make sure files are ingested in order of dependency
        return batches

    def write_files(self, *paths: Union[Path, str], raise_on_duplicate_file=False, skip_complete=True,
                    collision_manager='ignore', batch_size=None, parts=None, halt_on_error=True,
                    dryrun=False, do_not_apply_constraints=False, test_one=False, batches_slc: slice = None,
                    debug=False, debug_time=False, debug_params=False, debug_plan=False, timeout=None,
                    replan='skip_except_first', runtime='pipelined',
                    ) -> Optional[pd.DataFrame]:
        """
        Read in the files given in `paths` to the database.
        `collision_manager` is the method with which the database deals with overwriting data.
        Values of `collision_manager` can be {'ignore', 'overwrite', 'track&flag'}.
        track&flag will have the same behaviour as ignore but places the overlapping data in its own node for later retrieval.
        :return
            statistics dataframe
        """
        if not halt_on_error:
            logging.warning(f"`halt_on_error` is set to False. This would allow files to read without first reading their dependent files. "
                            f"This is not recommended.")
        if not do_not_apply_constraints:
            self.apply_constraints()
        batches = self.files_to_batches(*paths, batch_size=batch_size, batches_slc=batches_slc, parts=parts)
        if not batches:
            return
        elapsed_times = []
        stats = []
        timestamps = []
        if dryrun:
            logging.info(f"Dryrun: will not write to database. However, reading is permitted")
        _, path, slc, part = batches[0]
        desc_size = max(len(f'{path}[{slc.start}:{slc.stop}:{part}]') for filetype, path, slc, part in batches)
        first = f"{path}[{slc.start}:{slc.stop}:{part}]"
        bar = tqdm(batches, desc=first.ljust(desc_size))
        if debug or debug_time:
            with open('debug-timestamp.log', 'w') as f:
                pass
        out_batches = []
        for i, (filetype, path, slc, part) in enumerate(bar):
            bar.set_description(f'{path}[{slc.start}:{slc.stop}:{part}]')
            try:
                if raise_on_duplicate_file:
                    if len(self.graph.execute('MATCH (f:File {fname: $fname}) return count(f)', fname=path.name)) != 0:
                        raise FileExistsError(f"{path.name} exists in the DB and raise_on_duplicate_file=True")
                if skip_complete:
                    if self.file_batch_is_complete(filetype, path, slc, part):
                        continue
                with self.write_cypher(collision_manager) as query:
                    filetype.read(self.rootdir, path, slc, part)
                    cypher, params = query.render_query(as_lines=True)
                    uuid = f"//{uuid4()}"
                    guarantee, added_data = self.mark_batch_complete_query(path.name, slc, part, filetype.length(self.rootdir / path, part), filetype.parts)
                    if replan == 'skip_except_first':
                        replan_cypher = f'CYPHER replan=skip'
                        replan = 'skip'
                    else:
                        replan_cypher = f'CYPHER replan={replan}'
                    lines = [uuid, f'CYPHER runtime={runtime}', replan_cypher]  # runtime is to avoid neo4j bug with pipelines: https://github.com/neo4j/neo4j/issues/12441
                    if debug_plan:
                        lines.append('PROFILE')
                    lines += cypher[:-1] + ['WITH time0', guarantee, cypher[-1]]
                    params.update(added_data)
                    cypher = '\n'.join(lines)
                if debug:
                    with open('debug-query.log', 'w') as f:
                        f.write(cypher)
                    if debug_params:
                        with open('debug-params.log', 'w') as f:
                            f.write(self.graph.output_for_debug(**params, arrow=False, cmdline=False, silent=True))
                start = time.time()
                timed_out = False
                if not dryrun:
                    try:
                        self.graph.execute('MATCH (n) remove n._query_hash')
                        self.graph.execute('MATCH ()-[r]-() remove r._query_hash')
                        results = self.graph.execute(cypher, **params)
                        stats.append(results.stats())
                        timestamp = results.evaluate()
                        successful = True
                    except (ConnectionError, RuntimeError, IndexError, ConnectionResetError) as e:
                        is_running = True
                        timed_out = False
                        while is_running:
                            is_running = self.graph.execute("CALL dbms.listQueries() YIELD query WHERE query STARTS WITH $uuid return count(*)", uuid=uuid).evaluate()
                            time.sleep(1)
                            bar.set_description(f'{path}[{slc.start}:{slc.stop}:{part}]')
                            if timeout is not None:
                                if time.time() - start > timeout:
                                    timed_out = True
                                    break
                        r = self.graph.execute('MATCH (f:File {fname: $fname}) return timestamp()', fname=path.name)
                        timestamp = r.evaluate()
                        successful = timestamp is not None
                        stats.append(r.stats())
                        if not successful and halt_on_error:
                            if timed_out:
                                raise ConnectionTimeOutError(f"Connection timed out whilst writing {path}{slc.start}:{slc.stop}:{part} and it failed") from e
                            else:
                                raise ConnectionError(f"{path} could not be written to the database see neo4j logs for more details") from e
                    finally:
                        self.graph.execute('MATCH (n) remove n._query_hash')
                        self.graph.execute('MATCH ()-[r]-() remove r._query_hash')
                    if timestamp is None:
                        logging.warning(f"This query terminated early due to either an empty input table/data or "
                                        f"a match within the query returned no matches or it timed-out. "
                                        f"Adjust your `.read` method and query to allow for empty tables/data")
                    timestamps.append(timestamp)
                    if successful:
                        elapsed_times.append(time.time() - start)
                    else:
                        elapsed_times.append(-1)
                    out_batches.append((filetype, path, slc, part))
                    if debug or debug_time:
                        with open('debug-timestamp.log', 'a') as f:
                            f.write(str(elapsed_times[-1]) + '\n')
            except (ClientError, DatabaseError, FileExistsError, ConnectionError) as e:
                logging.exception('ClientError:', exc_info=True)
                if halt_on_error:
                    raise e
                print(e)
            if test_one:
                batches = batches[i:i+1]
                logging.info(f"Writing terminated because `test_one=True`")
                break
        if len(out_batches) and not dryrun:
            df = pd.DataFrame(stats)
            df['timestamp'] = timestamps
            df['elapsed_time'] = elapsed_times
            _, df['fname'], slcs, parts = zip(*out_batches)
            df['batch_start'], df['batch_end'] = zip(*[(i.start, i.stop) for i in slcs])
            df['part'] = parts
        elif dryrun:
            pass
            # df = pd.DataFrame(columns=['elapsed_time', 'fname', 'batch_start', 'batch_end', 'part'])
            # df['elapsed_time'] = elapsed_times
        else:
            df = pd.DataFrame(columns=['timestamp', 'elapsed_time', 'fname', 'batch_start', 'batch_end'])
        if dryrun:
            return
        return df.set_index(['fname', 'batch_start', 'batch_end', 'part'])

    def find_files(self, *filetype_names, skip_complete_files=True):
        filetype_names = [x+'_file' if not x.endswith('_file') else x for x in filetype_names]
        if any(f for f in filetype_names if f not in [i.singular_name for i in self.filetypes]):
            raise KeyError(f"Some or all of the filetype_names are not understood. "
                           f"Allowed names are: {[i.singular_name.replace('_file', '') for i in self.filetypes]}")
        filelist = []
        if len(filetype_names) == 0:
            filetypes = self.filetypes
        else:
            filetypes = [f for f in self.filetypes if f.singular_name in filetype_names]

        for filetype in filetypes:
            filelist += sorted([i for i in filetype.match_files(self.rootdir, self.graph)], key=lambda f: f.name)
        if skip_complete_files:
            extant_fnames = self.get_complete_files() if skip_complete_files else []
            filtered_filelist = [i for i in filelist if str(i.name) not in extant_fnames]
        else:
            filtered_filelist = filelist
        diff = len(filelist) - len(filtered_filelist)
        if diff:
            print(f'Skipping {diff} complete files (use skip_complete_files=False to go over them again)')
        return [f for f in tqdm(filtered_filelist, desc='getting MOS files') if File.check_mos(f)]

    def write_directory(self, *filetype_names, collision_manager='ignore', skip_extant_files=True, halt_on_error=False,
                        batch_size=None, parts=None, dryrun=False, **kwargs) -> pd.DataFrame:
        filtered_filelist = self.find_files(*filetype_names, skip_extant_files=skip_extant_files)
        return self.write_files(*filtered_filelist, collision_manager=collision_manager, halt_on_error=halt_on_error,
                                dryrun=dryrun, batch_size=batch_size, parts=parts, **kwargs)

    def _validate_one_required(self, hierarchy_name, timestamp=None):
        hierarchy = self.singular_hierarchies[hierarchy_name]
        if hierarchy.is_template:
            return
        subclasses = get_all_subclasses(hierarchy)
        parents = [h for h in hierarchy.parents]
        children = [h for h in hierarchy.children]
        dfs = []
        for relation in parents+children:
            if isinstance(relation, Multiple):
                mn, mx = relation.minnumber, relation.maxnumber
                b = relation.node.__name__
                reflected = relation.one2one
                relation_idname = relation.relation_idname
            else:
                mn, mx = 1, 1
                b = relation.__name__
                reflected = False
                relation_idname = None
            mn = 0 if mn is None else mn
            mx = 9999999 if mx is None else mx
            a = hierarchy.__name__
            if relation in parents:
                arrows = ['<-', '-']
                reldef = [a, b]
            else:
                arrows = ['-', '->']
                reldef = [b, a]
            relid1 = f'WHERE r1.relation_id = "{relation_idname}"' if relation_idname else ''
            relid2 = f'WHERE r2.relation_id = "{relation_idname}"' if relation_idname else ''
            subclass_exclusions = ' WHERE ' + ' AND '.join([f'NOT n:{i.__name__}' for i in subclasses]) if len(subclasses) else ''
            q = f"""
            MATCH (n:{a}){subclass_exclusions}
            OPTIONAL MATCH (n){arrows[0]}[r1:is_required_by]{arrows[1]}(m:{b}) {relid1}
            WITH n, count(r1) as forwardCount
            OPTIONAL MATCH (m:{b}){arrows[0]}[r2:is_required_by]{arrows[1]}(n) {relid2}
            WITH n, forwardCount, count(r2) as reflCount           
            WITH *
            WHERE forwardCount < {mn} OR forwardCount > {mx} OR (reflCount <> forwardCount AND {reflected})
            RETURN "{reldef[0]}" as child, "{reldef[1]}" as parent, {mn} as mn, {mx} as mx, {reflected} as reflected, id(n), forwardCount, reflCount            
            """
            result = self.graph.neograph.run(q).to_data_frame()
            dfs.append(result)
        try:
            df = pd.concat(dfs)
            return df, df.groupby(['child', 'parent']).apply(len).astype(int)
        except KeyError:
            return df, pd.Series()
        except ValueError:
            return pd.DataFrame(), pd.Series()

    def _validate_no_duplicate_relation_ordering(self, time=None):
        q = """
        MATCH (a)-[r1]->(b)<-[r2]-(a)
        WHERE TYPE(r1) = TYPE(r2) AND r1.order <> r2.order
        WITH a, b, apoc.coll.union(COLLECT(r1), COLLECT(r2))[1..] AS rs
        RETURN DISTINCT labels(a), id(a), labels(b), id(b), count(rs)+1
        """
        return self.graph.neograph.run(q).to_data_frame()

    def _validate_no_duplicate_relationships(self, timestamp=None):
        q = """
        MATCH (a)-[r1]->(b)<-[r2]-(a)
        WHERE TYPE(r1) = TYPE(r2) AND PROPERTIES(r1) = PROPERTIES(r2)
        WITH a, b, apoc.coll.union(COLLECT(r1), COLLECT(r2))[1..] AS rs
        RETURN DISTINCT labels(a), id(a), labels(b), id(b), count(rs)+1
        """
        return self.graph.neograph.run(q).to_data_frame()

    def _validate_rel_unique_ids(self, *hiers):
        """
        For each hierarchy which defines an identifier builder based on parent hierarchies,
        check that the parents+self is unique in the db since this is not guaranteed by neo4j indexes
        """
        instance_duplications = []
        for h in tqdm(hiers or list(self.singular_hierarchies.keys()), desc='checking for schema uniqueness violations'):
            hierarchy = self.singular_hierarchies[h]
            if hierarchy.has_rel_identity() and not hierarchy.is_template:
                id_factors = [f for f in hierarchy.factors if f in hierarchy.identifier_builder]
                id_parents = [p if isinstance(p, Multiple) else OneOf(p) for p in hierarchy.parents]
                id_children = [p if isinstance(p, Multiple) else OneOf(p) for p in hierarchy.children]
                id_parents = {p.node.__name__: (p.minnumber, p.maxnumber) for p in id_parents if p.name in hierarchy.identifier_builder}
                id_children = {p.node.__name__: (p.minnumber, p.maxnumber) for p in id_children if p.name in hierarchy.identifier_builder}
                parent_nodes = [f"<-[rp{i}:is_required_by]-(p{i}:{p})" for i, p in enumerate(id_parents)]
                parent_nodes2 = [f"<-[rp{i}2:is_required_by]-(p{i})" for i, p in enumerate(id_parents)]
                child_nodes = [f"-[rc{i}:is_required_by]->(c{i}:{c})" for i, c in enumerate(id_children)]
                child_nodes2 = [f"-[rc{i}2:is_required_by]->(c{i})" for i, c in enumerate(id_children)]
                nodes = parent_nodes+child_nodes
                nodes2 = parent_nodes2+child_nodes2
                node = f"(n:{hierarchy.__name__})"
                node2 = f"(m:{hierarchy.__name__})"
                first = f"{node}{nodes[0]}"
                first2 = f"{node2}{nodes2[0]}"
                others = ','.join([f"(n){p}" for p in nodes[1:]])
                others2 = ','.join([f"(m){p}" for p in nodes2[1:]])
                parent_vars = ','.join([f"p{i}" for i in range(len(parent_nodes))]+[f"c{i}" for i in range(len(child_nodes))])
                rel_vars = ','.join([f"rp{i}.order" for i in range(len(parent_nodes))]+[f"rc{i}.order" for i in range(len(child_nodes))])
                rel_vars2 = ','.join([f"rp{i}2.order" for i in range(len(parent_nodes2))]+[f"rc{i}2.order" for i in range(len(child_nodes2))])
                property_eq = ' and '.join([f"m.{f} = n.{f}" for f in id_factors])
                if property_eq:
                    property_eq = f' and {property_eq}'
                if others:
                    first = f"{first},{others}"
                if others2:
                    first2 = f"{first2},{others2}"
                q = f"""
                MATCH {first}
                WITH n, {parent_vars}, [{rel_vars}] as coll1
                OPTIONAL MATCH {first2}
                where n <> m {property_eq} and [{rel_vars2}] = coll1
                WITH n, count(m) as nduplicates, count(*) as cnt_rels
                where nduplicates = cnt_rels  // if even 1 is different, then it's not a duplicate
                return labels(n)[0] as type, id(n)
                """
                duplicates = self.graph.neograph.run(q).to_data_frame()
                if len(duplicates):
                    instance_duplications.append(duplicates)
        try:
            instance_duplications = pd.concat(instance_duplications)
            class_duplications = instance_duplications.groupby('type').apply(len)
            return instance_duplications, class_duplications
        except ValueError:
            return [], []


    def _validate_expected_number_schema(self):
        schema_violations = []
        label_instances = []
        for h in tqdm(list(self.singular_hierarchies.keys()), desc='checking for schema parent/child # violations'):
            df = self._validate_one_required(h)
            if df is not None:
                df, label_instance = df
                schema_violations.append(df)
                label_instances.append(label_instance)
        schema_violations = pd.concat(schema_violations)
        label_instances = pd.concat(label_instances)
        return schema_violations, label_instances

    def _validate_relation_constraints(self):
        qs = []
        for h in tqdm(list(self.singular_hierarchies.keys()), desc='checking for rel constraint schema violations'):
            hierarchy = self.singular_hierarchies[h]
            if not hierarchy.is_template:
                for rel in hierarchy.parents + hierarchy.children:
                    if isinstance(rel, Multiple):
                        for constraint in rel.constrain:
                            q = f"""
                            MATCH (h:{hierarchy.__name__})--(r:{rel.node.__name__})--(c:{constraint.__name__})
                            WITH h, r, size(collect(c)) as size
                            where size <> 1
                            return h as node, r as dependent, "{constraint.__name__}" as constraint, size as ndifferent_constaints
                            """
                            qs.append(self.graph.neograph.run(q).to_data_frame())
        return qs

    def _validate_all_nodes_created_with_file(self):
        q = dedent("""match (f:File) 
        with collect(distinct f._dbcreated) as dbs
        match (n) where not n.`_dbcreated` in dbs
        with labels(n) as labels, count(n) as cnt, collect(id(n)) as ids
        RETURN labels, cnt, ids""")
        return self.graph.execute(q).to_data_frame()

    def validate(self, timestamp: int = None, file_paths: List[Path]=None):
        level = logging.getLogger().level
        logging.getLogger().setLevel(logging.INFO)
        missing = []
        with pd.option_context('display.max_columns', None):
            if file_paths is not None:
                logging.info(f"Checking all requested files are present...")
                extant = {path: fname for path, fname in self.graph.execute(f'MATCH (n:File) return n.path as path, n.fname as fname')}
                missing = [str(path) for path in file_paths if path in extant]
                if missing:
                    logging.warning('fail')
                    logging.warning(f"\tMissing {len(missing)} files from db")
                    logging.warning(f"\tMissing the following files in the database:")
                    logging.warning(f"{missing}")
                else:
                    logging.info('pass')
            logging.info(f"Checking no duplicate relationships...")
            duplicates = self._validate_no_duplicate_relationships(timestamp)
            if len(duplicates):
                logging.warning(f"fail")
                logging.warning(f'\tThere are {len(duplicates)} duplicate relations')
                logging.warning(f"\n{duplicates}")
                logging.warning(f"\tThis indicates a weaveio logic issue. Raise with maintainer.")
            else:
                logging.info(f"pass")
            logging.info(f"Checking no duplicate relationships with different ordering...")
            duplicates = self._validate_no_duplicate_relation_ordering(timestamp)
            if len(duplicates):
                logging.warning(f"fail")
                logging.warning(f'\tThere are {len(duplicates)} relations with different orderings')
                logging.warning(f"\n{duplicates}")
                logging.warning(f"\tThis indicates a weaveio logic issue. Raise with maintainer.")
            else:
                logging.info(f"pass")
            logging.info(f"Checking expected number of parent/child relations from schema...")
            schema_violations, label_instances = self._validate_expected_number_schema()
            if len(schema_violations):
                logging.warning(f"fail")
                logging.warning(f'\tThere are {len(schema_violations)} violations of expected relationship number ({len(label_instances)} class relations)')
                logging.warning(f"\n{schema_violations}")
                logging.warning(f"\n{label_instances}")
                logging.warning(f"\tThis indicates incorrect file read logic on user defined read functions. Check read function.")
            else:
                logging.info(f"pass")
            logging.info(f"Checking unique relationships based on connected nodes are unique...")
            unique_rel_id_duplicates, label_rel_id_duplicates = self._validate_rel_unique_ids()
            if len(unique_rel_id_duplicates):
                logging.warning(f"fail")
                logging.warning(f"\tThere are {len(unique_rel_id_duplicates)} unique rel id violations ({len(label_rel_id_duplicates)} class definitions)")
                logging.warning(f"\n{unique_rel_id_duplicates}")
                logging.warning(f"\n{label_rel_id_duplicates}")
                logging.warning(f"\tThis indicates a weaveio logic issue. Raise with maintainer.")
            else:
                logging.info(f"pass")
            if len(missing) + len(duplicates) + len(schema_violations) + len(unique_rel_id_duplicates) == 0:
                logging.info('Database schema is valid!')
            else:
                logging.warning(f"Database schema is invalid")
            logging.getLogger().setLevel(level)
            return duplicates, (schema_violations, label_instances), unique_rel_id_duplicates

    def is_product(self, factor_name, hierarchy_name):
        return self.singular_name(factor_name) in self.singular_hierarchies[self.singular_name(hierarchy_name)].products

    def is_factor_name(self, name):
        if name in self.factor_hierarchies:
            return True
        try:
            name = self.singular_name(name)
            return self.is_singular_factor(name) or self.is_singular_idname(name)
        except KeyError:
            return False

    def is_singular_idname(self, value):
        return self.is_singular_name(value) and value.split('.')[-1] in self.singular_idnames

    def is_plural_idname(self, value):
        return self.is_plural_name(value) and value.split('.')[-1] in self.plural_idnames

    def is_plural_factor(self, value):
        return self.is_plural_name(value) and value.split('.')[-1] in self.plural_factors

    def is_singular_factor(self, value):
        return self.is_singular_name(value) and value.split('.')[-1] in self.singular_factors

    def class_name(self, name):
        if isinstance(name, type):
            return name.__name__
        else:
            return self.singular_hierarchies[self.singular_name(name)].__name__

    def plural_name(self, name):
        if isinstance(name, type):
            if issubclass(name, Hierarchy):
                return name.plural_name
            else:
                raise TypeError(f'{name} is not a weaveio object or string')
        if name in self.class_hierarchies:
            return self.class_hierarchies[name].plural_name
        name = name.lower()
        if self.is_plural_name(name):
            return name
        if self.is_singular_name(name):
            try:
                return self.singular_factors[name]
            except KeyError:
                try:
                    return self.singular_idnames[name]
                except KeyError:
                    try:
                        return self.relative_names[name]
                    except KeyError:
                        return self.singular_hierarchies[name].plural_name
        if '.' in name:
            pattern = name.lower().split('.')
            if any(map(self.is_plural_name, pattern)):
                return name
            return '.'.join(pattern[:-1] + [self.plural_name(pattern[-1])])
        return make_plural(name)

    def singular_name(self, name):
        if isinstance(name, type):
            if issubclass(name, Hierarchy):
                return name.singular_name
            else:
                raise TypeError(f'{name} is not a weaveio object or string')
        if name in self.class_hierarchies:
            return self.class_hierarchies[name].singular_name
        name = name.lower()
        if self.is_singular_name(name):
            return name
        if self.is_plural_name(name):
            try:
                return self.plural_factors[name]
            except KeyError:
                try:
                    return self.plural_idnames[name]
                except KeyError:
                    try:
                        return self.plural_relative_names[name]
                    except KeyError:
                        return self.plural_hierarchies[name].singular_name
        if '.' in name:
            pattern = name.lower().split('.')
            return '.'.join([self.singular_name(p) for p in pattern])
        return make_singular(name)

    def is_valid_name(self, name):
        if not isinstance(name, str):
            return False
        name = name.lower()
        if self.is_plural_name(name) or self.is_singular_name(name):
            return True
        if isinstance(name, str):
            pattern = name.split('.')
            if len(pattern) == 1:
                return self.is_plural_name(name) or self.is_singular_name(name)
            return all(self.is_valid_name(p) for p in pattern)
        return False

    def is_plural_name(self, name):
        """
        Returns True if name is a plural name of a hierarchy
        e.g. spectra is plural for Spectrum
        """
        if not isinstance(name, str):
            return False
        name = name.lower()
        if name in self.plural_hierarchies or name in self.plural_factors or\
                   name in self.plural_idnames or name in self.plural_relative_names:
            return True
        pattern = name.split('.')
        if len(pattern) == 1:
            return name in self.plural_hierarchies or name in self.plural_factors or\
                   name in self.plural_idnames or name in self.plural_relative_names
        return all(self.is_plural_name(n) for n in pattern)

    def is_singular_name(self, name):
        if not isinstance(name, str):
            return False
        name = name.lower()
        if name in self.singular_hierarchies or name in self.singular_factors or \
           name in self.singular_idnames or name in self.relative_names:
            return True
        pattern = name.split('.')
        if len(pattern) == 1:
            return name in self.singular_hierarchies or name in self.singular_factors or \
                   name in self.singular_idnames or name in self.relative_names
        return all(self.is_singular_name(n) for n in pattern)

    def __getitem__(self, address):
        if not isinstance(address, str):
            raise KeyError(f"{self} cannot be indexed by integers.")
        try:
            return self.query.__getitem__(address)
        except (UserError, KeyError) as e:
            self.query.raise_error_with_suggestions(address, e)

    def __getattr__(self, item):
        if item in ['keys', 'items', 'values']:
            # catch fastapi jsonable_encoder requests for serialization
            raise AttributeError(f"{self} has no attribute {item}")
        if item.startswith('_'):
            # query attributes are not allowed to start with '_'
            # if we are here, then that means __getattribute__ has failed and we should fail here
            # like a normal object
            raise AttributeError(f"{self} has no attribute {item}")
        try:
            return self.query.__getattr__(item)
        except (UserError, KeyError) as e:
            self.query.raise_error_with_suggestions(item, e)

    def plot_relations(self, i=-1, show_hdus=True, fname='relations', format='pdf', include=None):
        graph = self.relation_graphs[i]
        if not show_hdus:
            graph = nx.subgraph_view(graph, lambda n: not issubclass(n, HDU))  # True to get rid of templated
        # G = nx.subgraph_view(graph, filter_edge=lambda a, b: graph.edges[a, b]['style'] !=  'dotted')
        G = graph
        if include is not None:
            include = [self.singular_hierarchies[i] for i in include]
            include_list = include.copy()
            include_list += [a for i in include for a in nx.ancestors(G, i)]
            include_list += [d for i in include for d in nx.descendants(G, i)]
            G = nx.subgraph_view(G, lambda n: n in include_list)
        plot_graph(G, fname, format)

    def find_names(self, guess, n=10):
        """
        Finds the closest name to guess in the hierarchy. Returns the top n matches.
        """
        sources = [self.singular_hierarchies, self.plural_hierarchies, self.singular_factors, self.plural_factors]
        suggestions = {i for source in sources for i in source}
        inorder = sorted(suggestions, key=lambda x: textdistance.jaro_winkler(guess, x), reverse=True)
        return inorder[0:n]


    def _autosuggest(self, a, relative_to=None):
        a = self.singular_name(a)
        if relative_to is not None:
            relative_to = self.singular_name(relative_to)
        distance, distance_reverse = textdistance.jaro_winkler, True
        suggestions = []

        for h_singular_name, h in self.singular_hierarchies.items():
            newsuggestions = []
            if relative_to is not None:
                try:
                    self.paths_to_hierarchy(relative_to, h_singular_name, singular=True)
                except NetworkXNoPath:
                    hier = h.singular_name
                    factors = h.products_and_factors
                except NodeNotFound:
                    continue
                else:
                    hier = h.plural_name
                    factors = [self.plural_name(f) for f in h.products_and_factors]
                newsuggestions.append(hier)
                newsuggestions += factors
            else:
                newsuggestions += [h.singular_name, h.plural_name] + h.products_and_factors
                newsuggestions += [self.plural_name(f) for f in h.products_and_factors]
            try:
                newsuggestions.index(a)
            except ValueError:
                suggestions += newsuggestions
            else:
                return [a]
        inorder = sorted(list(set(suggestions)), key=lambda x: distance(a, x), reverse=distance_reverse)
        return inorder[:3]

    def autosuggest(self, a: str, relative_to: str = None, exception: Exception = None):
        if isinstance(exception, CardinalityError):
            l = [self.plural_name(a)]
        else:
            l = self._autosuggest(a, relative_to)
        string = '\n'.join([f'{i}. {s}' for i, s in enumerate(l, start=1)])
        msg = f"did you mean one of:\n{string}"
        raise exception.__class__(f"{str(exception.args[0])}\n{msg}") from exception

    def _get_by_neo4j_id(self, neoid, plural_name):
        from weaveio.readquery.functions import neo4j_id
        q = self.query.__getattr__(plural_name)
        return q[neo4j_id(q) == int(neoid)]


from collections import defaultdict
from copy import copy
from typing import List, Tuple, Dict
from pathlib import Path
import warnings

import networkx as nx
import pandas as pd
from astropy.table import Table

from .utilities import mask_infs, remove_successive_duplicate_lines, dtype_conversion
from .digraph import HashedDiGraph, plot_graph, add_start, add_traversal, add_filter, add_aggregation, add_operation, add_return, add_unwind, subgraph_view, get_above_state_traversal_graph, node_dependencies, add_node_reference
from .statements import StartingMatch, Traversal, NullStatement, Operation, GetItem, AssignToVariable, DirectFilter, CopyAndFilter, Aggregate, Return, Unwind, GetProduct, UnionTraversal, ApplyToList


class ParserError(Exception):
    pass


class DeadEndException(Exception):
    pass


def traverse(graph, start=None, end=None, done=None, ordering=None, views=None):
    """
    traverse the traversal_graph with backtracking
    """
    if views is None:
        dag = subgraph_view(graph, excluded_edge_type='wrt')
        backwards_graph = subgraph_view(graph, only_edge_type='wrt')
        traversal_graph = subgraph_view(dag, excluded_edge_type='dep')
        dep_graph = subgraph_view(graph, only_edge_type='dep')
        views = (dag, backwards_graph, traversal_graph, dep_graph)
    else:
        dag, backwards_graph, traversal_graph, dep_graph = views
    if start is None or end is None:
        naive_ordering = list(nx.topological_sort(dag))
        if start is None:
            start = naive_ordering[0]  # get top node
        if end is None:
            end = naive_ordering[-1]
    if ordering is None:
        ordering = [start]
    else:
        ordering = list(ordering)
    node = start
    done = set() if done is None else set(done)  # stores wrt edges and visited nodes
    while True:
        dependencies = dep_graph.predecessors(node)
        if not all(dep in done for dep in dependencies):
            raise DeadEndException
        options = [b for b in backwards_graph.successors(node) if (node, b) not in done]  # must do wrt first
        if not options:
            options = [o for o in traversal_graph.successors(node) if o not in done]   # where to go next?
            option_edges = [graph.edges[(node, o)] for o in options]
            op_options = []
            for o, e in zip(options, option_edges):
                if e.get('type', '') == 'operation' and e.get('single', False):
                    if all(dep in done for dep in dep_graph.predecessors(o)):
                        op_options.append(o)
            # do all operations first since they can't be wrong!
            options = op_options + [o for o in options if o not in op_options]
        if not options:
            # if you cant go anywhere and you're not done, then this recursive path is bad
            if node != end:
                raise DeadEndException
            else:
                return ordering, done
        elif len(options) == 1:
            # if there is only one option, go there... obviously
            edge = (node, options[0])
            if edge in done:
                # recursive path is bad if you have to go over the same wrt edge more than once
                raise DeadEndException
            elif graph.edges[edge]['type'] == 'wrt':
                done.add(edge)  # only add edges for wrt
                # and now also reset the branch listed in `ordering[iwrt:inode] so it can be traversed again`
                # only reset the nodes/edges which have other dep-paths outside the branch
                most_recent_mention = len(ordering) - ordering[::-1].index(options[0]) - 1
                this_branch = ordering[most_recent_mention:]
                still_todo = set(graph.nodes) - done - set(this_branch)
                restricted_dag = nx.subgraph_view(dag, lambda n: n != node)
                # TODO: also include this edge we are on now... might be needed later somehow
                for n in set(this_branch[:-1]):
                    if any(nx.has_path(restricted_dag, n, a) for a in still_todo):
                        done -= {n}
                        for wrt_edge in backwards_graph.successors(n):
                            done -= {(n, wrt_edge)}
            done.add(node)
            node = options[0]
            ordering.append(node)
        else:
            # open up recursive paths from each available option
            # this is such a greedy algorithm
            for option in options:
                try:
                    new_done = done.copy()
                    new_ordering = ordering.copy()
                    new_ordering.append(option)
                    new_done.add(node)
                    new_done.add(option)
                    ordering, new_done = traverse(graph, option, end, tuple(new_done), tuple(new_ordering), views)
                    done.update(new_done)
                    node = ordering[-1]
                    break
                except DeadEndException:
                    pass  # try another option
            else:
                raise DeadEndException  # all options exhausted, entire recursive path is bad


def verify_traversal(graph, traversal_order):
    edges = list(zip(traversal_order[:-1], traversal_order[1:]))
    if any(graph.edges[e]['type'] == 'dep' for e in edges):
        raise ParserError(f"Some dep edges where traversed. This is a bug")
    semi_dag = subgraph_view(graph, excluded_edge_type='dep')
    if set(semi_dag.edges) != set(edges):
        raise ParserError(f"Not all edges were traversed. This is a bug")
    done = set()
    for n in traversal_order:
        if n not in done:
            if not all(dep in done for dep in node_dependencies(graph, n)):
                raise ParserError(f"node {n} does not have all its dependencies satisfied. This is a bug")
            done.add(n)


def verify(graph):
    """
    Check that edges and nodes are allowed:
        - There is only one output node and one input node (no hanging nodes)
        - There is a path from input->output
        - can only aggregate to a parent
        - There are no cyclic dependencies in the dag
        - can only use an aggregation when it's wrt is a parent
        - all operations must be aggregated
        - Multiple inputs into a node should comprise:
            all deps that are aggregated
            one other (can be anything)
        - For an agg node, there is only one wrt
        - You can have > 1 inputs when they are ops

        - Multiple outputs from a node:
            no more than one out-path should be unaggregated in the end
            (i.e. there should only be one path from start-output which contains no aggregations)
    """
    dag = subgraph_view(graph, excluded_edge_type='wrt')
    traversal = subgraph_view(dag, excluded_edge_type='dep')
    if not nx.is_arborescence(traversal):
        raise ParserError(f"Invalid query: The DAG for this query is not a directed tree with max 1 parent per node")
    starts = [n for n in dag.nodes if dag.in_degree(n) == 0]
    ends = [n for n in dag.nodes if dag.out_degree(n) == 0]
    if len(starts) != 1:
        raise ParserError("Only one input node is allowed")
    if len(ends) > 1:
        raise ParserError("Only one output node is allowed")
    if not ends:
        raise ParserError("An output node is required")
    backwards = subgraph_view(graph, only_edge_type='wrt')
    without_agg = subgraph_view(traversal, excluded_edge_type='aggr')
    main_paths = nx.all_simple_paths(without_agg, starts[0], ends[0])
    try:
        next(main_paths)
        next(main_paths)
    except StopIteration:
        pass
    else:
        # there can be 0 in the case where the output is itself an aggregation
        raise ParserError(f"There can only be at maximum one path from {starts[0]} to {ends[0]} that is not aggregated")
    if not nx.is_directed_acyclic_graph(dag):
        raise ParserError(f"There are cyclical dependencies")
    if not nx.has_path(dag, starts[0], ends[0]):
        raise ParserError(f"There must be a path from {starts[0]} to {ends[0]}")
    for agg, wrt in backwards.edges:
        if not nx.has_path(graph, wrt, agg):
            raise ParserError(f"{wrt} must be a parent of {agg} in order to aggregate")
        for node in dag.successors(agg):
            if not nx.has_path(graph, wrt, node):
                raise ParserError(f"{node} can an only use what is aggregated above it. failure on {agg} (parent={wrt})")
    for node in graph.nodes:
        inputs = [graph.edges[i]['type'] for i in graph.in_edges(node)]
        inputs = [i for i in inputs if i != 'wrt']
        outputs = [graph.edges[i]['type'] for i in graph.out_edges(node)]
        if sum(o == 'wrt' for o in outputs) > 1:
            raise ParserError(f"Cannot put > 1 wrt paths as output from an aggregation: {node}")
        outputs = [o for o in outputs if o != 'wrt']
        nfilters = sum(i == 'filter' for i in inputs)
        ntraversals = sum(i == 'traversal' for i in inputs)
        ndeps = sum(i == 'dep' for i in inputs)
        nops = sum(i == 'operation' for i in inputs)
        naggs = sum(i == 'aggr' for i in inputs)
        nreturns = sum(i == 'return' for i in inputs)
        if naggs > 1:
            raise ParserError(f"Cannot aggregate more than one node at a time: {node}")
        elif naggs:
            if not all(o in ['dep', 'operation', 'aggr'] for o in outputs):
                raise ParserError(f"Can only use aggregations as a dependency/operation/aggregation afterwards {node}")
        if nfilters > 2:
            raise ParserError(f"Can only have one filter input: {node}")
        elif nfilters:
            if ntraversals + nops + naggs > 0:
                raise ParserError(f"A filter can only take dependencies not traversals/ops/aggregations: {node}")
        if ntraversals > 2:
            raise ParserError(f"Can only have one traversal input: {node}")
        elif ntraversals:
            if len(inputs) > 2:
                raise ParserError(f"Can only traverse with one input or one input and one unwind: {node}")
        if nops > 1:
            raise ParserError(f"Can only have one op input: {node}")
        elif nops:
            try:
                if graph.edges[list(graph.out_edges(node))[0]]['type'] not in ['aggr', 'operation']:
                    raise ParserError(f"All operations must be aggregated back at some point: {node}")
            except IndexError:
                raise ParserError(f"All operations must be aggregated back at some point: {node}")
            if ntraversals + naggs + nfilters > 1:
                raise ParserError(f"Can only have dependencies as input for an operation: {node}")


def merge_attribute_forks(graph):
    """
    Given a node in the graph find simple (no external deps) successor branches which wrt back to the same node.
    Merge these parallel branches into one serial branch
    """


def simplify_graph(graph: HashedDiGraph):
    return graph


class QueryGraph:
    """
    Rules of adding nodes/edges:
    Traversal:
        Can only traverse to another hierarchy object if there is a path between them
        Always increases/maintains cardinality
    Aggregation:
        You can only aggregate back to a predecessor of a node (the parent)
        Nodes which require another aggregation node must share the same parent as just defined above

    Golden rule:
        dependencies of a node must share an explicit parent node
        this basically says that you can only compare nodes which have the same parents

    optimisations:
        If the graph is duplicated in multiple positions, attempt to not redo effort
        For instance, if you traverse and then agg+filter back to a parent and the traverse the same path
        again after filtering, then the aggregation is changed to conserve the required data and the duplicated traversal is removed

    """

    def __init__(self):
        self.G = HashedDiGraph()
        self.start = add_start(self.G, 'data')
        self.variable_names = defaultdict(int)
        self.dag_G = nx.subgraph_view(self.G, filter_edge=lambda a, b: self.G.edges[(a, b)]['type'] != 'wrt')  # type: nx.DiGraph
        self.backwards_G = nx.subgraph_view(self.G, filter_edge=lambda a, b: self.G.edges[(a, b)]['type'] == 'wrt')  # type: nx.DiGraph
        self.traversal_G = nx.subgraph_view(self.G, filter_edge=lambda a, b: self.G.edges[(a, b)]['type'] != 'dep')  # type: nx.DiGraph
        self.parameters = {}
        self.groupbys = {}

    def statements(self, node):
        if node is None:
            G = self.G
        else:
            G = self.restricted(node)
        return {d['statement']: (a, b) for a, b, d in G.edges(data=True) if 'statement' in d}

    def latest_shared_ancestor(self, *nodes):
        if all(n == nodes[0] for n in nodes):
            return nodes[0]
        return sorted(set.intersection(*[self.above_state(n, no_wrt=True) for n in nodes]), key=lambda n: len(nx.ancestors(self.traversal_G, n)))[-1]

    def latest_object_node(self, *nodes):
        """
        a and b have a shared ancestor
        Scenarios:
            a = single; b = single -> shared
            a = single; b = plural -> choose originating object of b
            a = plural; b = single -> choose ordering object of a
            a = plural; b = plural -> disallowed [at least one must be aggregated back]
        """
        cardinals = []
        for node in nodes:
            try:
                cardinal = next(self.backwards_G.successors(node))
            except StopIteration:
                cardinal = node
            cardinals.append(cardinal)

        shared = self.latest_shared_ancestor(*cardinals)
        is_shared = [shared == c for c in cardinals]
        if all(is_shared):
            return shared
        if sum(is_shared) == len(is_shared) - 1:
            for c, i in zip(cardinals, is_shared):
                if not i:
                    return c
        else:
            raise ParserError(f"One of {nodes} must be a parent of the other. {shared} not in {nodes} ")

    def is_singular_branch(self, a, b):
        path = nx.shortest_path(self.above_graph, b, a)
        return all(self.above_graph.edges[(b, a)].get('single', True) for b, a in zip(path[:-1], path[1:]))

    def is_singular_branch_relative_to_splits(self, node):
        path = nx.shortest_path(self.above_graph, node, self.start)
        for b, a in zip(path[:-1], path[1:]):
            edge = self.above_graph.edges[(b, a)]
            if edge.get('split_node', False):
                return True
            if not edge.get('single', True):
                return False
        return True  # if we get to this point, we're at the start node and all edges are single



    @property
    def above_graph(self):
        return get_above_state_traversal_graph(self.G)

    def node_holds_type(self, node, *types):
        return any(d['type'] in types for a, b, d in self.G.in_edges(node, data=True))

    def get_host_nodes_for_operation(self, op_node) -> set:
        """
        Find nodes which are necessary to construct `op_node`
        traverses until it encounters a traversal/filter/aggr then it stops
        An operation can have a single path (just scalar ops) or a branching mess (combining ops),
        """
        necessary = set()
        for input_edge in self.dag_G.in_edges(op_node, data=True):
            input = input_edge[0]
            if self.node_holds_type(input, 'traversal', 'filter', 'aggr'):
                necessary.add(input)
            else:
                necessary |= self.get_host_nodes_for_operation(input)
        return necessary

    def above_state(self, node, no_wrt=False):
        states = nx.descendants(self.above_graph, node)
        states.add(node)
        if not no_wrt:
            for n in states.copy():
                aggregates = self.backwards_G.predecessors(n)
                for aggr in aggregates:
                    edge = list(self.traversal_G.in_edges(aggr))[0]
                    if isinstance(self.G.edges[edge]['statement'], NullStatement):
                        # this is a fake aggregation so get all required nodes as well
                        # TODO: do not do this if the branch is not SINGLE!
                        op_stuff = self.get_host_nodes_for_operation(aggr)
                        states |= op_stuff
                    else:
                        states.add(aggr)
        return states

    def get_variable_name(self, name):
        name = name.lower()
        new_name = f'{name}{self.variable_names[name]}'
        self.variable_names[name] += 1
        return new_name

    def export(self, fname, result_node=None, directory=None, highlight_nodes=None,
               highlight_edges=None, ftype='pdf'):
        return plot_graph(self.restricted(result_node), highlight_nodes, highlight_edges).\
            render(fname, directory, format=ftype)

    def export_slideshow(self, dirname, ordering, result_node=None):
        Path(dirname).mkdir(parents=True, exist_ok=True)
        for i, (a, b) in enumerate(zip(ordering[:-1], ordering[1:])):
            self.export(f'{i}', result_node, dirname, [b], [(a, b)], ftype='png')

    def add_start_node(self, node_type, unwound=None):
        parent_node = self.start
        statement = StartingMatch(node_type, unwound, self)
        return add_traversal(self.G, parent_node, statement, unwound=unwound)

    def add_traversal(self, parent_node, paths: List[str], end_node_type: str, single=False, unwound=None):
        statement = Traversal(self.G.nodes[parent_node]['variables'][0], end_node_type, paths, unwound, self)
        return add_traversal(self.G, parent_node, statement, single=single, unwound=unwound)

    def fold_to_cardinal(self, parent_node, wrt=None):
        """
        Adds a fake aggregation such that a node can be used as a dependency later.
        For operation chains, you follow the traversal route backwards.
        If the node
        """
        agg = next(self.backwards_G.successors(parent_node), None)
        if wrt is None:
            if agg is not None:
                return parent_node  # if its already aggregated then do nothing
            else:
                path = nx.shortest_path(self.traversal_G, self.start, parent_node)[::-1]
                for b, a in zip(path[:-1], path[1:]):
                    if not self.G.edges[(a, b)]['single']:
                        wrt = b
                        break
                else:
                    raise ParserError
        elif agg == wrt:
            return parent_node
        if wrt == parent_node:
            return parent_node
        return self.add_aggregation(parent_node, wrt, 'aggr')

    def add_scalar_operation(self, parent_node, op_format_string, op_name, parameters=None) -> Tuple:
        """
        A scalar operation is one which takes only one input and returns one output argument
        the input can be one of [object, operation, aggregation]
        """
        if any(d['type'] == 'aggr' for a, b, d in self.G.in_edges(parent_node, data=True)):
            wrt = next(self.backwards_G.successors(parent_node))
            return self.add_combining_operation(op_format_string, op_name, parent_node, wrt=wrt, parameters=parameters)
        statement = Operation(self.G.nodes[parent_node]['variables'][0], [], op_format_string, op_name, self, parameters)
        return add_operation(self.G, parent_node, [], statement), parent_node

    def add_combining_operation(self, op_format_string, op_name, *nodes, wrt=None, parameters=None) -> Tuple:
        """
        A combining operation is one which takes multiple inputs and returns one output
        Operations should be inline (no variables) for as long as possible.
        This is so they can be used in match where statements
        """
        # if this is combiner operation, then we do everything with respect to the nearest ancestor
        dependency_nodes = [self.fold_to_cardinal(d) for d in nodes]  # fold back when combining
        if wrt is None:
            wrt = self.latest_object_node(*dependency_nodes)
        deps = [self.G.nodes[d]['variables'][0] for d in dependency_nodes]
        statement = Operation(deps[0], deps[1:], op_format_string, op_name, self, parameters)
        return add_operation(self.G, wrt, dependency_nodes, statement), wrt

    def add_getitem(self, parent_node, item, which=0):
        statement = GetItem(self.G.nodes[parent_node]['variables'][which], item, self)
        return add_operation(self.G, parent_node, [], statement)

    def add_getproduct(self, parent_node, item):
        statement = GetProduct(self.G.nodes[parent_node]['variables'][0], item, self)
        return add_operation(self.G, parent_node, [], statement)

    def assign_to_variable(self, parent_node, only_if_op=False):
        if only_if_op and any(d['type'] in ['operation', 'aggr'] for a, b, d in self.G.in_edges(parent_node, data=True)):
            stmt = AssignToVariable(self.G.nodes[parent_node]['variables'][0], self)
            return add_operation(self.G, parent_node, [], stmt)
        return parent_node

    def add_generic_aggregation(self, parent_node, wrt_node, op_format_string, op_name,
                                remove_infs=None, expected_dtype=None, input_dtype=None):
        if expected_dtype is not None:
            op_format_string = dtype_conversion(input_dtype, expected_dtype, op_format_string, '{0}')
        if remove_infs:
            op_format_string = op_format_string.format(mask_infs('{0}'))
        if wrt_node == parent_node or op_name == 'aggr':
            statement = NullStatement(self.G.nodes[parent_node]['variables'] + [wrt_node], self)
        else:
            if wrt_node not in nx.ancestors(self.dag_G, parent_node):
                raise SyntaxError(f"{parent_node} cannot be aggregated to {wrt_node} ({wrt_node} is not an ancestor of {parent_node})")
            statement = Aggregate(self.G.nodes[parent_node]['variables'][0], wrt_node, op_format_string, op_name, self)
        previous = next(self.backwards_G.successors(parent_node), parent_node)
        dependencies = [] if previous == parent_node else [parent_node]
        return add_aggregation(self.G, previous, wrt_node, statement, dependencies=dependencies)

    def add_aggregation(self, parent_node, wrt_node, op, remove_infs=None, expected_dtype=None, input_dtype=None, distinct=False, nulls=False):
        dst = 'distinct ' if distinct else ''
        if nulls:
            aggr = f"[i in collect({dst}[{{0}}]) | i[0]]"
        else:
            aggr = f"{op}({dst}{{0}})"
        return self.add_generic_aggregation(parent_node, wrt_node, aggr, op, remove_infs, expected_dtype, input_dtype)

    def add_predicate_aggregation(self, parent, wrt_node, op_name):
        op_format_string = f'{op_name}(x in collect({{0}}) where toBoolean(x))'
        return self.add_generic_aggregation(parent, wrt_node, op_format_string, op_name)

    def add_filter(self, parent_node, predicate_node, direct=False, force_single=False, split_node=False):
        wrt = self.latest_shared_ancestor(parent_node, predicate_node)
        predicate_node = self.fold_to_cardinal(predicate_node, wrt)  # put everything back to most recent shared ancestor
        if not nx.has_path(self.G, predicate_node, parent_node):
            raise SyntaxError(f"{parent_node} cannot be filtered by {predicate_node} since there is no direct path between them")
        predicate = self.G.nodes[predicate_node]['variables'][0]
        if direct:
            FilterClass = DirectFilter
        else:
            FilterClass = CopyAndFilter
        statement = FilterClass(self.G.nodes[parent_node]['variables'][0], predicate, self)
        return add_filter(self.G, parent_node, [predicate_node], statement, force_single, split_node=split_node)

    def add_apply_to_list(self, parent_node, list_variable, apply_function, filter_function,
                          *dependencies, put_null_in_empty=False, parameters=None):
        dependencies = [self.fold_to_cardinal(d) for d in dependencies]
        statement = ApplyToList(list_variable, [self.G.nodes[d]['variables'][0] for d in dependencies],
                                apply_function, filter_function, self, put_null_in_empty, parameters=parameters)
        return add_operation(self.G, parent_node, dependencies, statement), statement.output_variables[0]

    def add_unwind_parameter(self, wrt_node, to_unwind, *dependencies, parameters=None):
        statement = Unwind(wrt_node, to_unwind, 'unwound', self, parameters=parameters)
        return add_unwind(self.G, wrt_node, statement, *dependencies)

    def collect_or_nots(self, index_node, other_nodes, force_plurals, treat_equal, already_aggregated_to_index):
        """
        Collect `other_node` with respect to the shared common ancestor of `index_node` and `other_node`.
        If other_node is above the index, fold back to cardinal node
        if not, fold back to shared ancestor
        """
        # successors = [next(self.backwards_G.successors(n, None)) for n in other_nodes]
        # shareds = [self.latest_shared_ancestor(index_node, n) for n in other_nodes]
        if already_aggregated_to_index:
            return other_nodes
        results = []
        for other_node, force_plural in zip(other_nodes, force_plurals):
            successor = next(self.backwards_G.successors(other_node), None)
            if not force_plural and successor == index_node:
                # if its not already aggregated to the index
                if treat_equal:
                    return other_nodes
                results.append(other_node)
                continue
            shared = self.latest_shared_ancestor(index_node, other_node)
            if force_plural:
                if treat_equal:
                    return [self.add_aggregation(o, shared, 'collect') for o in other_nodes]
                results.append(self.add_aggregation(other_node, shared, 'collect'))
            else:
                if successor == shared:
                    if treat_equal:
                        return other_nodes
                    results.append(other_node)
                else:
                    how = 'aggr' if self.is_singular_branch(shared, other_node) else 'collect'
                    if treat_equal:
                        return [self.add_aggregation(o, shared, how) for o in other_nodes]
                    results.append(self.add_aggregation(other_node, shared, how))
        return results

    def add_results_table(self, index_node, column_nodes, force_plurals: List[bool], dropna=None,
                          treat_equal=False, already_aggregated_to_index=False):
        # fold back column data into the index node
        column_nodes = self.collect_or_nots(index_node, column_nodes, force_plurals, treat_equal, already_aggregated_to_index) # fold back when combining
        deps = [self.G.nodes[d]['variables'][0] for d in column_nodes]
        try:
            vs = self.G.nodes[index_node]['variables'][0]
        except IndexError:
            vs = None
        if dropna is None:
            try:
                dropna = self.G.nodes[index_node]['variables'][0]
            except IndexError:
                dropna = None
        else:
            dropna = self.G.nodes[dropna]['variables'][0]
        statement = Return(deps, vs, dropna, self)
        collected = [list(self.G.in_edges(c, data='type'))[0][-1] == 'collect' for c in column_nodes]
        return add_return(self.G, index_node, column_nodes, statement), collected

    def add_scalar_results_row(self, *column_nodes):
        """data already folded back"""
        deps = [self.G.nodes[d]['variables'][0] for d in column_nodes]
        statement = Return(deps, None, [], self)
        return add_return(self.G, self.start, column_nodes, statement)

    def add_groupby(self, query):
        name = self.add_parameter('<placeholder>', 'group')
        self.groupbys[name] = query
        return name

    def add_parameter(self, value, name=None):
        if isinstance(value, pd.DataFrame):
            value = Table.from_pandas(value)
        if isinstance(value, Table):
            # limit column multidimensional sizes to 100
            allowed_cols = [c for c in value.colnames if sum(value[c].shape) / len(value) <= 100]
            if allowed_cols != value.colnames:
                warnings.warn(f"Columns {set(value.colnames) - set(allowed_cols)} were dropped due to size limitations (>=100)")
                value = value[allowed_cols]
        # if value in self.parameters.values():
        #     return [k for k, v in self.parameters.items() if v is value][0]
        for k, v in self.parameters.items():
            if value is v:
                return k
        name = f'${name}' if name is not None else '$'
        varname = self.get_variable_name(name)
        self.parameters[varname] = value
        return varname

    def add_previous_reference(self, parent_node, node_to_reference):
        vars = self.G.nodes[node_to_reference]['variables']
        return add_node_reference(self.G, NullStatement(vars, self), parent_node, node_to_reference)

    def restricted(self, result_node=None) -> HashedDiGraph:
        if result_node is None:
            return nx.subgraph_view(self.G)
        return nx.subgraph_view(self.G, lambda n: nx.has_path(self.dag_G, n, result_node))

    def dependency_parameters(self, result_node):
        ps = set()
        for statement in self.statements(result_node):
            if statement.parameters is not None:
                ps |= set(statement.parameters)
        return {k: v for k, v in self.parameters.items() if k in ps}

    def traverse_query(self, result_node=None, simplify=True):
        graph = self.restricted(result_node)
        # verify(graph)
        if simplify:
            graph = simplify_graph(graph)
        return traverse(graph)[0]

    def verify_traversal(self, goal, ordering, simplify=True):
        graph = self.restricted(goal)
        if simplify:
            graph = simplify_graph(graph)
        return verify_traversal(graph, ordering)

    def cypher_lines(self, result, no_cache=False):
        try:
            cypher = self.G.nodes[result]['cypher']
            ordering = self.G.nodes[result]['ordering']
        except KeyError:
            ordering = self.traverse_query(result)
            self.verify_traversal(result, ordering)
            statements = []
            for i, e in enumerate(zip(ordering[:-1], ordering[1:])):
                try:
                    statement = self.G.edges[e]['statement'].make_cypher(ordering[:i+1])
                    if statement is not None:
                        statements.append(statement)
                except KeyError:
                    pass
            cypher = remove_successive_duplicate_lines(statements)
            if not no_cache:
                self.G.nodes[result]['cypher'] = cypher
                self.G.nodes[result]['ordering'] = ordering
        return copy(cypher)

    def node_is_null_statement(self, node):
        if self.node_holds_type(node, 'aggr'):
            return any(isinstance(d.get('statement', None), NullStatement) for _, _, d in self.G.in_edges(node, data=True))
        return False

    def get_unwind_variables(self, until_node):
        variables = {d['statement'].parameter: d['statement'].output for *e, d in self.restricted(until_node).edges(data=True) if isinstance(d.get('statement', None), Unwind)}
        return {p: v for k, v in variables.items() for p in sorted(self.parameters.keys(), key=len, reverse=True) if p in k}



#!/usr/bin/env python

import inspect
import os
import subprocess
import sys
from textwrap import dedent

try:
    import weaveio
    from weaveio import *
except ImportError as e:
    raise Exception(f"weaveio is not installed. Please install it and activate its environment") from e


def shell_command(command, capture_output=True, echo=True, dryrun=False):
    if echo:
        print(f"running `{command}`")
    if dryrun:
        return
    if capture_output:
        shell_command = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        stdout, stderr = shell_command.communicate()
        shell_command.wait(5)
        if shell_command.returncode != 0:
            raise Exception(f"{command} failed with return code {shell_command.returncode}")
        return stdout.decode('utf-8').strip('\n')
    return subprocess.Popen(command, shell=True)


def get_default_args(func):
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }


def get_neo4j_version():
    return shell_command('echo "call dbms.components() yield name, versions, edition unwind versions as version return name, version, edition;" | cypher-shell -u weaveuser -p weavepassword | tail -n 1')


def get_apoc_version():
    return shell_command('echo "return apoc.version() AS output;" | cypher-shell -u weaveuser -p weavepassword | tail -n 1')


def get_python_version():
    import sys
    return f"Python version: {sys.version}"


def version(args):
    print(f"python version: {get_python_version()}")
    print(f"weaveio version: {weaveio.__version__}")
    print(f"neo4j version: {get_neo4j_version()}")
    print(f"apoc version: {get_apoc_version()}")


def console(args):
    import IPython
    data = Data(rootdir=args.rootdir, host=args.host, port=args.port, user=args.user, password=args.password, dbname=args.dbname)
    header = dedent(f"""WeaveIO console version {weaveio.__version__}
data = Data(rootdir='{data.rootdir}', host='{data.host}', port='{data.port}', user='{data.user}', dbname='{data.dbname}')
run `%matplotlib` to allow interactive plots""")
    IPython.embed(header=header, matplotlib='auto', colors='Linux')

def jupyter_start(args):
    print('Remember to forward the port to the notebook server')
    if args.port is None:
        shell_command('jupyter notebook --no-browser --ip 0.0.0.0 --allow-root', capture_output=False, dryrun=args.dryrun)
    else:
        shell_command(f'jupyter notebook --no-browser --ip 0.0.0.0 --port {args.port} --allow-root', capture_output=False, dryrun=args.dryrun)
    print('Notebook server started')
    print('To stop the notebook server, run weaveio jupyter --stop')


def jupyter_stop(args):
    if args.port is None:
        print('killing all weaveio jupyter servers at the following ports')
        print(shell_command("pgrep -f 'weaveio/bin/jupyter-notebook'", capture_output=True, echo=False, dryrun=args.dryrun))
        input('press enter to continue\n>>>')
        shell_command("pkill -f 'weaveio/bin/jupyter-notebook'", capture_output=False, dryrun=args.dryrun)


def jupyter(args):
    if args.list:
        print('weaveio jupyter servers at the following ports:')
        try:
            print(shell_command("pgrep -f 'weaveio/bin/jupyter-notebook'", capture_output=True, echo=False, dryrun=args.dryrun))
        except Exception:
            pass
    elif args.stop:
        return jupyter_stop(args)
    else:
        return jupyter_start(args)


def upgrade(args):
    return shell_command(f"{str(Path(sys.executable).parents[0] / 'pip install -U weaveio --no-cache-dir')}", capture_output=False, dryrun=args.dryrun)

def add_data(args):
    raise NotImplementedError(f"This feature has not yet been written")

def main():
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument('--dryrun', action='store_true', help='do not actually run any commands')
    sub_parsers = parser.add_subparsers(help='sub-command help')

    parser_version = sub_parsers.add_parser('version', help='display weaveio version')
    parser_version.set_defaults(func=version)
    parser_console = sub_parsers.add_parser('console', help='start weaveio ipython console')
    parser_console.add_argument('--dbname', default='production', help='root directory of data')
    parser_console.add_argument('--rootdir', default=None, help='root directory of data')
    parser_console.add_argument('--host', default=None, help='neo4j host')
    parser_console.add_argument('--port', default=None, help='neo4j port')
    parser_console.add_argument('--user', default=None, help='neo4j user')
    parser_console.add_argument('--password', default=None, help='neo4j password')
    parser_console.set_defaults(func=console)
    parser_jupyter = sub_parsers.add_parser('jupyter', help='start a jupyter notebook in the weaveio environment in this directory')
    parser_jupyter.add_argument('--port', help='port to bind to, default is port found in jupyter configuration')
    parser_jupyter.add_argument('--stop', action='store_true', help='stop weaveio jupyter server at selected port or all of them if not specified')
    parser_jupyter.add_argument('--list', action='store_true', help='list weaveio jupyter servers')
    parser_jupyter.set_defaults(func=jupyter)
    parser_upgrade = sub_parsers.add_parser('upgrade', help='upgrade weaveio through pip')
    parser_upgrade.set_defaults(func=upgrade)
    parser_add_data = sub_parsers.add_parser('add_data', help='add data to weaveio')
    parser_add_data.set_defaults(func=add_data)

    args = parser.parse_args()
    if args.dryrun:
        print(f"dryrun mode: {args}")
    if 'func' not in args:
        parser.print_help()
        exit(1)
    args.func(args)

if __name__ == '__main__':
    main()


import pytest
from hypothesis import given
import hypothesis.strategies as st

import numpy as np
from hypothesis.strategies import composite

import weaveio.basequery.dissociated
from weaveio.writequery import CypherQuery, merge_node, match_node, unwind, collect, groupby, CypherData, merge_relationship, set_version
from weaveio.graph import Graph


PREFIX = "WITH *, timestamp() as time0"

def run_test_query(query, graph: Graph):
    cypher, params = query.render_query()
    return graph.execute(cypher, **params)


def test_context_compatibility():
    query = CypherQuery()
    graph = Graph(host='host.docker.internal')
    with graph:
        with query:
            assert CypherQuery.get_context() is query
            assert Graph.get_context() is graph


def test_merge_node(write_database: Graph):
    for i in range(2):
        with CypherQuery() as query:
            merge_node(labels=['A', 'B'], identproperties={'a': 1, 'b': 2, 'c': [1,2,3]})
        run_test_query(query, write_database)
    t = write_database.execute('MATCH (n) return labels(n) as labels, n.a, n.b, n.c').to_data_frame()
    assert len(t) == 1
    row = t.iloc[0]
    assert sorted(row['labels']) == ['A', 'B']
    assert row['n.a'] == 1
    assert row['n.b'] == 2
    assert row['n.c'] == [1,2,3]


def test_match_node(write_database):
    with CypherQuery() as query:
        merge_node(labels=['A', 'B'], identproperties={'a': 1, 'b': 2, 'c': [1, 2, 3]})
        node = match_node(labels=['A', 'B'], properties={'a': 1, 'b': 2, 'c': [1, 2, 3]})
        query.returns(a=node['a'], b=node['b'], c=node['c'])
    result = run_test_query(query, write_database)
    t = result.to_data_frame()
    assert len(t) == 1
    row = t.iloc[0]
    assert row["a"] == 1
    assert row["b"] == 2
    assert row["c"] == [1,2,3]

baseproperty = st.text() | st.floats(allow_nan=True, allow_infinity=True) | st.integers()
property = baseproperty | st.lists(baseproperty)

@composite
def node(draw, labels=st.lists(st.text()), properties=st.dictionaries(st.text(), property), identproperties=baseproperty):
    return (draw(labels), draw(identproperties), draw(properties))

@composite
def relation(draw, reltype=st.text(), properties=st.dictionaries(st.text(), property), identproperties=baseproperty):
    return (draw(reltype), draw(identproperties), draw(properties))


@pytest.mark.parametrize('collision_manager', ['track&flag', 'ignore', 'overwrite'])
@pytest.mark.parametrize('different_parents', [True, False])
@pytest.mark.parametrize('different_identproperties', [True, False])
@pytest.mark.parametrize('different_properties', [True, False])
@pytest.mark.parametrize('different_reltype', [True, False])
@pytest.mark.parametrize('different_relidentproperties', [True, False])
@pytest.mark.parametrize('different_relproperties', [True, False])
@given(parents=st.lists(st.tuples(node(), relation()), min_size=0, max_size=5, unique=True))
class TestDependentMergeWithOneParent:
    @pytest.fixture(autouse=True)
    def database(self, write_database):
        """setup the parent nodes in the database first"""

    @pytest.fixture(autouse=True)
    def after_first(self):
        """Run the first command (i.e. the first merge to setup the test)"""
        # do something
        return self.database

    @pytest.fixture(autouse=True)
    def after_second(self):
        """Run the second command (i.e. the merge that should have some expected result, which is tested for)"""
        # do something
        self.first
        return self.first

    def test_merged_node_or_nodes_exist_as_expected(self):
        """Standard cypher query to check that all properties etc exist"""
        assert False

    def test_node_collisions_exist_if_necessary(self):
        assert False

    def test_rel_collisions_exist_if_necessary(self):
        assert False





def test_merge_many2one_with_one(write_database):
    with CypherQuery() as query:
        parent = merge_node(labels=['A', 'B'], identproperties={'a': 1, 'b': 2, 'c': [1, 2, 3]})
        child = merge_node(['C'], {'p': 1}, parents={parent: ['req', {'a': 1, 'b': 2}]})
        result = run_test_query(query, write_database)
    t = write_database.execute('MATCH (n: C)<-[r]-(p) return labels(n) as labels, n.a, n.b, n.c, p.p, type(r), r.a, r.b').to_data_frame()
    assert len(t) == 1
    row = t[0]
    assert row['n.a'] == 1
    assert row['n.b'] == 2
    assert row['n.c'] == [1,2,3]
    assert row['p.p'] == 1
    assert row['r.a'] == 1
    assert row['r.b'] == 2
    assert row['type(r)'] == 'req'
    assert row['labels(n)'] == ['C']


def test_merge_many2one_with_mixture():
    with CypherQuery() as query:
        parent1 = match_node(labels=['A', 'B'], properties={'a': 1})
        parent2 = match_node(labels=['A', 'B', 'other'], properties={'a': 2})
        child = merge_node(['C'], {'p': 1}, parents={parent1: 'req1', parent2: 'req2'}, versioned_label='other')
    cypher = query.render_query()[0]
    expected = [PREFIX,
        "MATCH (b0:A:B {a: 1})",
        "MATCH (other0:A:B:Other {a: 2})",
        "WITH *, [[b0, 'req1', {order: 0}]] as bspec0",
        "WITH *, [[other0, 'req2', {order: 0}]] as otherspec0",
        "WITH *, bspec0+otherspec0 as specs0",
        "CALL custom.multimerge(specs0, ['C'], {p: 1}, {dbcreated: time0}, {dbcreated: time0}) YIELD child as c0",
        "CALL custom.version(specs0, c0, ['Other'], 'version') YIELD version as version0",
        "RETURN time0"
                ]
    expected = '\n'.join(expected)
    assert cypher == expected


def test_merge_many2one_with_mixture_run(procedure_tag, write_database):
    with CypherQuery() as query:
        parent1 = merge_node(labels=['A', 'B'], properties={'a': 1})
        parent2 = merge_node(labels=['A', 'B', 'other'], properties={'a': 2})
        child = merge_node(['C'], {'p': 1}, parents={parent1: 'req1', parent2: 'req2'}, versioned_labels=['other'])
    cypher = query.render_query(procedure_tag)[0]
    write_database.neograph.run(cypher)
    result = write_database.neograph.run('MATCH (n: C) MATCH (p: A)-[]->(n) WITH n, collect(p.a) as a return n.p, a').to_table()
    assert len(result) == 1
    row = result[0]
    assert row[0] == 1
    assert row[1] == [1, 2]


def test_getitem():
    with CypherQuery() as query:
        parent1 = merge_node(labels=['A', 'B'], properties={'a': 1, 'b': 2})
        query.returns(parent1['a'], parent1['b'])
    cypher = query.render_query()[0]
    assert cypher.split('\n')[-1] == "RETURN b0['a'], b0['b']"


@pytest.mark.parametrize('input_data', [1, [1,2], 'a', [1, 'a'], np.array([1,2])])
def test_input_data_retrievable(input_data):
    with CypherQuery() as query:
        data1 = CypherData(input_data, 'mydata')
        data2 = CypherData(input_data, 'mydata')
        query.returns(data1, data2)
    cypher = query.render_query()[0]
    assert cypher == PREFIX + '\nRETURN $mydata0, $mydata1'


def test_unwind_input_list():
    with CypherQuery() as query:
        data = CypherData([1,2,3], 'mydata')
        merge_node(['B'], {'b': 1})
        with unwind(data) as number:
            node = merge_node(['A'], {'a': number})
        nodes = collect(node)
        query.returns(nodes)
    cypher = query.render_query()[0]
    expected = [
        PREFIX,
        'MERGE (b0:B {b: 1}) ON CREATE SET b0.dbcreated = time0',
        'WITH * UNWIND $mydata0 as unwound_mydata0',
            'MERGE (a0:A {a: unwound_mydata0}) ON CREATE SET a0.dbcreated = time0',
        'WITH time0, b0, collect(a0) as as0',
        "RETURN as0"
    ]
    assert cypher == '\n'.join(expected)


def test_unwind_nested_contexts():
    with CypherQuery() as query:
        data = CypherData([1,2,3], 'mydata')
        merge_node(['B'], {'b': 1})
        with unwind(data) as number1:
            with unwind(data) as number2:
                node = merge_node(['A'], {'a': number1, 'b': number2})
            nodelist = collect(node)
        nodelistlist = collect(nodelist)
        query.returns(nodelistlist)
    cypher = query.render_query()[0]
    expected = [
        PREFIX,
        'MERGE (b0:B {b: 1}) ON CREATE SET b0.dbcreated = time0',
        'WITH * UNWIND $mydata0 as unwound_mydata0',
            'WITH * UNWIND $mydata0 as unwound_mydata1',
                'MERGE (a0:A {a: unwound_mydata0, b: unwound_mydata1}) ON CREATE SET a0.dbcreated = time0',
            'WITH time0, b0, unwound_mydata0, collect(a0) as as0',
        'WITH time0, b0, collect(as0) as ass0',
        "RETURN ass0"
    ]
    assert cypher == '\n'.join(expected)


def test_autoclose_unwind_with_no_variables_carried_over():
    with CypherQuery() as query:
        data = CypherData([1,2,3], 'mydata')
        merge_node(['B'], {'b': 1})
        with unwind(data) as number:
            node = merge_node(['A'], {'a': number})
    cypher = query.render_query()[0].split('\n')
    assert cypher[-2] == 'WITH time0, b0'


def test_closed_unwind_context_not_accessible():
    with CypherQuery() as query:
        data = CypherData([1,2,3], 'mydata')
        merge_node(['B'], {'b': 1})
        with unwind(data) as number:
            node = merge_node(['A'], {'a': number})
        with pytest.raises(ValueError):
            merge_node(['B'], {'b': number})


def test_unwind_that_returns_multiple_variables():
    with CypherQuery() as query:
        data = CypherData(['1', '2', '3'], 'mydata')
        with unwind(data, data) as (a, b):
            pass
    cypher = query.render_query()[0]
    expected = [
        PREFIX,
        "WITH *, apoc.coll.max([x in [$mydata0,$mydata0] | SIZE(x)])-1 as m",
        "UNWIND range(0, m) as i0 WITH *, $mydata0[i0] as unwound_mydata0, $mydata0[i0] as unwound_mydata1",
        "WITH time0",
        "RETURN time0"
    ]
    assert cypher == '\n'.join(expected)


def test_unwind_that_returns_multiple_variables_with_enumerate():
    with CypherQuery() as query:
        data = CypherData(['1', '2', '3'], 'mydata')
        with unwind(data, enumerated=True) as (d, i):
            pass
        indexes = collect(i)
    cypher = query.render_query()[0]
    expected = [
        PREFIX,
        "WITH *, apoc.coll.max([x in [$mydata0] | SIZE(x)])-1 as m",
        "UNWIND range(0, m) as i0 WITH *, $mydata0[i0] as unwound_mydata0",
        "WITH time0, collect(i0) as is0",
        "RETURN time0"
    ]
    assert cypher == '\n'.join(expected)



def test_unwind_collect_unwind_collect():
    with CypherQuery() as query:
        data = CypherData(['1', '2', '3'], 'mydata')
        with unwind(data, enumerated=True) as (d, i):
            pass
        ds = collect(d)
        with unwind(ds) as d:
            pass
        ds = collect(d)
    cypher = query.render_query()[0]
    expected = [
        PREFIX,
        "WITH *, apoc.coll.max([x in [$mydata0] | SIZE(x)])-1 as m",
        "UNWIND range(0, m) as i0 WITH *, $mydata0[i0] as unwound_mydata0",
        "WITH time0, collect(unwound_mydata0) as unwound_mydatas0",
        "WITH * UNWIND unwound_mydatas0 as unwound_unwound_mydatas0",
        "WITH time0, collect(unwound_unwound_mydatas0) as unwound_unwound_mydatass0",
        "RETURN time0"
    ]
    assert cypher == '\n'.join(expected)


def test_accessing_unwound_variable_after_unwinding_is_not_allowed():
    with CypherQuery() as query:
        data = CypherData(['1', '2', '3'], 'mydata')
        with unwind(data, enumerated=True) as (d, i):
            pass
        ds = collect(d)
        with unwind(ds) as d:
            pass
        with pytest.raises(ValueError):
            groupby(ds, 'property')


def test_groupby_makes_dict(procedure_tag, write_database):
    with CypherQuery() as query:
        data = CypherData(['1', '2', '3'], 'mydata')
        merge_node(['B'], {'b': 1})
        with unwind(data) as number:
            node = merge_node(['A'], {'a': number})
        nodes = collect(node)
        nodedict = groupby(nodes, 'a')
        query.returns(nodedict['1']['a'], nodedict['2']['a'], nodedict['3']['a'], nodedict['0']['a'])
    cypher, parameters = query.render_query(procedure_tag)
    result = write_database.neograph.run(cypher, parameters=parameters).to_table()
    assert result[0] == ('1', '2', '3', None)


def test_groupby_fails_if_input_is_not_collection():
    with CypherQuery() as query:
        data = CypherData(['1', '2', '3'], 'mydata')
        b = merge_node(['B'], {'b': 1})
        with unwind(data) as number:
            node = merge_node(['A'], {'a': number})
        with pytest.raises(TypeError):
            nodedict = groupby(b, 'b')


@pytest.fixture(scope='class')
def bigquery(write_database, procedure_tag):
    with CypherQuery() as query:
        checksums = CypherData(list(range(10)),'checksums')
        runids = CypherData(['runid1', 'runid2', 'runid3'], 'runids')
        with unwind(checksums, enumerated=True) as (checksum, i):
            fibretarget = merge_node(['target'], {'id': i})
        fibretargets = collect(fibretarget)
        with unwind(checksums, fibretargets) as (checksum, fibretarget):
            with unwind(runids) as runid:
                run = merge_node(['run'], {'id': runid})
                raw = merge_node(['raw'], {}, parents={run: 'req'}, versioned_labels=['run'])
                spec = merge_node(['SingleSpectrum'], {}, parents={raw: 'req', fibretarget: 'req'}, versioned_labels=['raw', 'target'])
            specs = collect(spec)
            stack = merge_node(['stack'], {'checksum': checksum}, parents={specs: 'req'})
        stacks = collect(stack)
        merge_node(['StackFile'], {'fname': 'fname'}, parents={stacks: 'req'})
        run = match_node(['run'], {'id': 'runid1'})
        raw = match_node(['raw'], {}, parents={run: 'req'})
        merge_node(['rawfile'], {'fname': 'rawfname'}, parents={raw: 'req'})
    cypher, data = query.render_query(procedure_tag)
    write_database.neograph.run(cypher, parameters=data)
    return cypher, data, write_database


@pytest.mark.usefixtures("bigquery")  # careful, there seems to be a bug with pytest and this database is not separate!
class TestBigQuery:
    def test_spec_has_run_and_raw(self, bigquery):
        cypher, data, database = bigquery
        result = database.neograph.run('match (s:SingleSpectrum) optional match (raw:Raw)-->(s)<--(t:Target) return s, raw, t').to_data_frame()
        assert len(result) == 30
        assert not weaveio.basequery.dissociated.any()

    @pytest.mark.parametrize('v,n,version', [('Run', 3, None), ('Raw', 3, 0), ('SingleSpectrum', 30, 0)])
    def test_raw_has_version0(self, v, n, version, bigquery):
        cypher, data, database = bigquery
        result = database.neograph.run(f'match (v:{v}) return v.version').to_data_frame()
        assert len(result) == n
        if version is not None:
            assert (result['v.version'] == version).all()
        else:
            assert weaveio.basequery.dissociated.all()

    def test_stack_has_3_single_spectra(self, bigquery):
        cypher, data, database = bigquery
        result = database.neograph.run(f'match (stack:Stack) optional match (stack)<--(single:SingleSpectrum) '
                                       f'with stack, collect(single) as singles return stack, singles').to_data_frame()
        assert len(result) == 10
        assert all(len(s) == 3 for s in result['singles'])

    def test_stackfile_has_all_stacks(self, bigquery):
        cypher, data, database = bigquery
        result = database.neograph.run(f'match (stackfile:StackFile) optional match (stackfile)<--(stack:Stack) '
                                       f'with stackfile, collect(stack.checksum) as stacks return stackfile, stacks').to_data_frame()
        assert len(result) == 1
        assert all(set(s) == set(range(10)) for s in result['stacks'])

    def test_rawfile_is_attached_to_a_single_raw(self, bigquery):
        cypher, data, database = bigquery
        result = database.neograph.run('match (f:Rawfile) optional match (run:Run)-->(raw:Raw)-->(f) return f, raw, run.id').to_data_frame()
        assert len(result) == 1
        assert result['run.id'][0] == 'runid1'


def test_merge_relationship():
    with CypherQuery() as query:
        a = merge_node(['a'], {})
        b = merge_node(['b'], {})
        merge_relationship(a, b, 'rel', {'prop': 'a'})
    cypher = query.render_query()[0]
    assert cypher == '\n'.join([PREFIX,
                                "MERGE (a0:A {}) ON CREATE SET a0.dbcreated = time0",
                                "MERGE (b0:B {}) ON CREATE SET b0.dbcreated = time0",
                                "MERGE (a0)-[rel0:rel {prop: 'a'}]->(b0) ON CREATE SET rel0.dbcreated = time0",
                                "RETURN time0"
                                ])


def test_merge_node_relationship_with_one():
    with CypherQuery() as query:
        a = merge_node(['a'], {})
        merge_dependent_node(['b'], {'prop': 'b'}, [(a, 'rel', {'prop': 'a'})])
    cypher = query.render_query()[0]
    assert cypher == '\n'.join([PREFIX,
                                "MERGE (a0:A {}) ON CREATE SET a0.dbcreated = time0",
                                "MERGE (a0)-[rel0:rel {prop: 'a'}]->(b0:B {prop: 'b'}) "
                                "ON CREATE SET rel0.dbcreated = time0, b0.dbcreated = time0",
                                "RETURN time0"
                                ])


def test_merge_node_relationship_with_two():
    with CypherQuery() as query:
        a = merge_node(['a'], {})
        b = merge_node(['b'], {})
        merge_dependent_node(['c'], {'prop': 'c'}, [(a, 'rel', {'prop': 'a'}),
                                                    (b, 'rel', {'prop': 'b'})])
    cypher = query.render_query()[0]
    assert cypher == '\n'.join([PREFIX,
                                "MERGE (a0:A {}) ON CREATE SET a0.dbcreated = time0",
                                "MERGE (b0:B {}) ON CREATE SET b0.dbcreated = time0",
                                "MERGE (a0)-[rel0:rel {prop: 'a'}]->(c0:C {prop: 'c'})<-[rel1:rel {prop: 'b'}]-(b0) "
                                "ON CREATE SET rel0.dbcreated = time0, c0.dbcreated = time0, rel1.dbcreated = time0",
                                "RETURN time0"
                                ])


def test_versioning(write_database):
    with CypherQuery() as query:
        a = merge_node(['a'], {})
        b = merge_node(['b'], {})
        c1 = merge_dependent_node(['c'], {'prop': 'c1'}, [(a, 'rel', {'prop': 'a'}),
                                                          (b, 'rel', {'prop': 'b'})])
        set_version([a, b], ['rel', 'rel'], 'C', c1, {})
        c2 = merge_dependent_node(['c'], {'prop': 'c2'}, [(a, 'rel', {'prop': 'a'}),
                                                          (b, 'rel', {'prop': 'b'})])
        set_version([a, b], ['rel', 'rel'], 'C', c2, {})
        c2dupl = merge_dependent_node(['c'], {'prop': 'c2'}, [(a, 'rel', {'prop': 'a'}),
                                                              (b, 'rel', {'prop': 'b'})])
        set_version([a, b], ['rel', 'rel'], 'C', c2dupl, {})
    cypher = query.render_query()[0]
    write_database.neograph.run(cypher)
    results = write_database.neograph.run('MATCH (c:C) return c.prop, c.version ORDER BY c.version').to_table()
    assert results[0] == ('c1', 0)
    assert results[1] == ('c2', 1)
    assert len(results) == 2



import pytest
from weaveio import *
from weaveio.readquery.results import Table


def test_aligned_attribute(data):
    red = data.l1single_spectra[data.l1single_spectra.camera == 'red']
    blue = red.adjunct
    aligned = align(red=red, blue=blue).camera
    table = aligned()
    assert isinstance(table, Table)
    assert table.colnames == ['red', 'blue']
    assert all((table['red'].data == 'red').filled(True))
    assert all((table['blue'].data == 'blue').filled(True))

def test_aligned_column(data):
    red = data.l1single_spectra[data.l1single_spectra.camera == 'red']
    blue = red.adjunct
    aligned = align(red=red, blue=blue)[['camera']]
    table = aligned()
    assert isinstance(table, Table)
    assert table.colnames == ['red_camera', 'blue_camera']
    assert all((table['red_camera'].data == 'red').filled(True))
    assert all((table['blue_camera'].data == 'blue').filled(True))

def test_aligning_to_table_by_from_others(data):
    red = data.l1single_spectra[data.l1single_spectra.camera == 'red']
    blue = red.adjunct
    aligned = align(red=red, blue=blue)
    avg = mean(aligned.l1stack_spectra.snr, wrt=aligned)
    snr = aligned.snr
    q = aligned[{'divide': avg / snr}, snr, avg]
    table = q()
    assert table.colnames == ['red_divide', 'red_snr', 'red_mean', 'blue_divide', 'blue_snr', 'blue_mean']
    np.testing.assert_allclose(table['red_divide'], table['red_mean'] / table['red_snr'])
    np.testing.assert_allclose(table['blue_divide'], table['blue_mean'] / table['blue_snr'])
